{Tera Term
 Copyright(C) 1994-1996 T. Teranishi
 All rights reserved.}

{TERATERM.EXE, IME interface}
unit IME;

interface

uses WinTypes, WinProcs, TTWinMan, TTCommon;

type
  TSendIMEMessage = function(HWin: HWnd; lParam: longint): word;
  TWINNLSEnableIME = function(HWin: HWnd; bEnable: bool): bool;

var
  HWINNLS: THandle;
  SendIMEMessage: TSendIMEMessage;
  WINNLSEnableIME: TWINNLSEnableIME;

function LoadWINNLS: bool;
procedure FreeWINNLS;
procedure SetConversionWindow(HWin: HWnd; X,Y: integer; Font: word);

const
  WM_IME_REPORT = $0280;
  IR_OPENCONVERT = $0120;
  IR_CLOSECONVERT = $0122;
  IR_STRING = $0140;

implementation

const
  MCW_Default = 0;
  MCW_Window = 2;

  IME_SetOpen = 4;
  IME_SetConversionWindow = 8;
  IME_SetConversionFont = 18;

type

  PIMEStruct = ^IMEStruct;
  IMEStruct = record
    fnc: word;
    wParam:    word;
    wCount:    word;
    dchSource: word;
    dchDest:   word;
    lParam1:   longint;
    lParam2:   longint;
    lParam3:   longint;
  end;

const 
  IdSendIMEMessage  = 6;
  IdWINNLSEnableIME = 16;

function LoadWINNLS: bool;
var
  Err: boolean;
begin
  if HWINNLS >= 32 then
  begin
    LoadWINNLS := true;
    exit;
  end
  else
    LoadWINNLS := false;

  HWINNLS := LoadLibrary('WINNLS.DLL');
  if HWINNLS < 32 then
  begin
    messagebox(HMainWin,'Non-Japanese Windows user may ignore this error.',
               'Tera Term: Error',mb_iconexclamation);
    WritePrivateProfileString('Tera Term','IME','off',ts.SetupFName);
    LoadDefaultSet(ts.SetupFName);
    exit;
  end;

  Err := false;
  @SendIMEMessage := GetProcAddress(HWINNLS, PChar(IdSendIMEMessage));
  if @SendIMEMessage=nil then Err := true;
  
  @WINNLSEnableIME := GetProcAddress(HWINNLS, PChar(IdWINNLSEnableIME));
  if @WINNLSEnableIME=nil then Err := true;

  if Err then
  begin
    FreeLibrary(HWINNLS);
    HWINNLS := 0;
  end
  else LoadWINNLS := true;

end;

procedure FreeWINNLS;
begin
  if HWINNLS >= 32 then
  begin
    {position of conv. window -> default}
    SetConversionWindow(HMainWin,-1,0,0);

    FreeLibrary(HWINNLS);
    HWINNLS := 0;
  end;
end;

procedure SetConversionWindow(HWin: HWnd; X,Y: integer; Font: word);
var
  HIME: THandle;
  lIMEParam: longint;
  PIME: PIMEStruct;
begin
  if HWINNLS < 32 then exit;

  HIME := GlobalAlloc(GHND,longint(SizeOf(IMEStruct)));

  {Set position of conversion window}
  PIME := PIMEStruct(GlobalLock(HIME));
  PIME^.fnc := IME_SetConversionWindow;
  if X<0 then PIME^.wParam := MCW_DEFAULT
         else PIME^.wParam := MCW_WINDOW;
  PIME^.lParam1 := MakeLong(X,Y);
  GlobalUnlock(HIME);
  lIMEParam := MakeLong(HIME,0);
  SendIMEMessage(HWin, lIMEParam);

  {Set font of conversion window}
  PIME := PIMEStruct(GlobalLock(HIME));
  PIME^.fnc := IME_SetConversionFont;
  PIME^.wParam := Font;
  GlobalUnlock(HIME);
  lIMEParam := MakeLong(HIME,0);
  SendIMEMessage(HWin, lIMEParam);

  GlobalFree(HIME);
end;

begin
  HWINNLS := 0;
end.