{Tera Term
 Copyright(C) 1994-1996 T. Teranishi
 All rights reserved.}

{TERATERM.EXE, keyboard routines}
unit Keyboard;

interface
{$I teraterm.inc}

{$IFDEF Delphi}
uses Messages, WinTypes, TTTypes, WinProcs, OWindows, Strings, WinDos,
     TTWinMan, TTCommon, TTSetup, TTLib;
{$ELSE}
uses WinTypes, TTTypes, WinProcs, WObjects, Strings, WinDos, Win31,
     TTWinMan, TTCommon, TTSetup, TTLib;
{$ENDIF}

const
  FuncKeyStrMax = 32;

var
  WaitKeyId, WaitHi: boolean;
  NewKeyId, NewKeyLen: integer;
  NewKeyStr: array[0..FuncKeyStrMax-1] of byte;
  FuncKeyStr: array[0..IdF20-IdF6, 0..FuncKeyStrMax-1] of char;
  FuncKeyLen: array[0..IdF20-IdF6] of integer;

  {keyboard status}
  PreviousKey: integer;
  AutoRepeatMode: boolean;
  AppliKeyMode, AppliCursorMode: boolean;
  OpenConv: BOOL; {for Win95J}

  {key code map}
  KeyMap: PKeyMap;

  {Character string sent by key pressiong}
  CodeCount, CodeLength: integer;
  Code: array[0..100] of Char;
  CodeConv: BOOL;

procedure SetKeyMap;
procedure ClearUserKey;
procedure DefineUserKey;
procedure KeyDown(HWin: HWnd; VKey, Count, Scan: WORD);
procedure KeyUp(VKey: WORD);
function ShiftKey: BOOL;
function ControlKey: BOOL;
function AltKey: BOOL;
procedure InitKeyboard;
procedure EndKeyboard;

implementation

procedure SetKeyMap;
var
 TempDir: array[0..fsDirectory] of char;
 TempName: array[0..fsFileName+fsExtension] of char;
begin
  if StrLen(ts.KeyCnfFN)=0 then exit;
  StrUpperJ(ts.KeyCnfFN);
  ExtractFileNameJ(ts.KeyCnfFN,TempName);
  ExtractDirNameJ(ts.KeyCnfFN,TempDir);

  if (StrIComp(TempDir,ts.HomeDir)<>0) or
     (StrIComp(TempName,'KEYBOARD.CNF')<>0) then
  begin
    if KeyMap=nil then New(KeyMap);
    if KeyMap<>nil then
    begin
      if LoadTTSet then
        ReadKeyboardCnf(ts.KeyCnfFN, KeyMap);
      FreeTTSet;
    end;
  end
  else if KeyMap<>nil then
  begin
    Dispose(KeyMap);
    KeyMap := nil;
  end;
end;

procedure ClearUserKey;
var
  i: integer;
begin
  for i := 0 to IdF20-IdF6 do
  begin
    FuncKeyLen[i] := 0;
  end; 
end;

procedure DefineUserKey;
begin
  if (NewKeyLen=0) or (NewKeyLen>FuncKeyStrMax) then exit;

  case NewKeyId of
    17..21: NewKeyId := NewKeyId-17;
    23..26: NewKeyId := NewKeyId-18;
    28..29: NewKeyId := NewKeyId-19;
    31..34: NewKeyId := NewKeyId-20;
  else
    exit;
  end;

  Move(NewKeyStr[0], FuncKeyStr[NewKeyId,0], NewKeyLen);
  FuncKeyLen[NewKeyId] := NewKeyLen;

end;

procedure KeyDown(HWin: HWnd; VKey, Count, Scan: WORD);
var
  Key : word;
  TM :TMsg;
  KeyState: array[0..255] of byte;
  i: integer;
begin
  {for Win95J}
  if OpenConv then exit;
  if (VKey=VK_Shift) or
     (VKey=VK_Control) or
     (VKey=VK_Menu) then exit;

  {debug mode}
  if ts.Debug and (VKey = VK_ESCAPE) and
     ShiftKey then
  begin
    messagebeep(0);
    DebugFlag := not DebugFlag;
    CodeCount := 0;
    PeekMessage(TM,HWin,WM_CHAR,WM_CHAR,PM_REMOVE);
    exit;
  end;

  if not AutoRepeatMode and (PreviousKey=VKey) then
  begin
    PeekMessage(TM,HWin,WM_CHAR,WM_CHAR,PM_REMOVE);
    exit;
  end;
  PreviousKey := VKey;

  if ShiftKey then Scan := Scan or $200;
  if ControlKey then Scan := Scan or $400;
  if AltKey then Scan := Scan or $800;

  CodeCount := Count;

  {Windows keycode -> Tera Term keycode}
  Key := GetKeyCode(KeyMap,Scan);

  CodeConv := true;
  if Key=0 then
    case VKey of
      VK_BACK:
        begin
          CodeLength := 1;
          if ControlKey then
            if ts.BSKey=IdDel then Code[0] := #$08
                              else Code[0] := #$7F
          else
            if ts.BSKey=IdDel then Code[0] := #$7F
                              else Code[0] := #$08;
        end;
      VK_RETURN: {CR Key}
        begin
          CodeLength := 1;
          Code[0] := #$0D;
        end;
      VK_SPACE:
        begin
          if ControlKey then
          begin
            CodeLength := 1;
            Code[0] := #0;
          end
	  else
            CodeLength := 0;
        end;
      else
        CodeLength := 0;
    end
  else
    if (IdF6<=Key) and (Key<=IdF20) and (FuncKeyLen[Key-IdF6]>0) then
    begin
      Move(FuncKeyStr[Key-IdF6,0], Code[0], FuncKeyLen[Key-IdF6]);
      CodeLength := FuncKeyLen[Key-IdF6];
      CodeConv := false;
    end
    else
      GetKeyStr(HMainWin,KeyMap,Key,AppliKeyMode,AppliCursorMode,
                Code,@CodeLength,@CodeConv);

  if CodeLength=0 then exit;

  if VKey=VK_NUMLOCK then
  begin
    {keep NumLock LED status}
    GetKeyboardState(TKeyboardState(KeyState));
    KeyState[VK_NUMLOCK] := KeyState[VK_NUMLOCK] xor 1;
    SetKeyboardState(TKeyboardState(KeyState));
  end;

  PeekMessage(TM,HWin,WM_CHAR,WM_CHAR,PM_REMOVE);

  if TalkStatus=IdTalkKeyb then
  begin
    if CodeConv then
      for i := 1 to  CodeCount do
      begin
        if ts.LocalEcho>0 then
          CommTextEcho(@cv,Code,CodeLength);
        CommTextOut(@cv,Code,CodeLength);
      end
    else
      for i := 1 to CodeCount do
      begin
        CommBinaryOut(@cv,Code,CodeLength);
        if ts.LocalEcho>0 then
          CommBinaryEcho(@cv,Code,CodeLength);
      end;
  end;

end;

procedure KeyUp(VKey: WORD);
begin
  if PreviousKey = VKey then PreviousKey := 0;
end;

function ShiftKey: BOOL;
begin
  ShiftKey := GetAsyncKeyState(VK_SHIFT) and $FF80 <> 0;
end;

function ControlKey: BOOL;
begin
  ControlKey := GetAsyncKeyState(VK_CONTROL) and $FF80 <> 0;
end;

function AltKey: BOOL;
begin
  AltKey := GetAsyncKeyState(VK_MENU) and $FF80 <> 0;
end;

procedure InitKeyboard;
begin
  KeyMap := nil;
  ClearUserKey;
  OpenConv := FALSE;
  PreviousKey := 0;
end;

procedure EndKeyboard;
begin
  if KeyMap <> nil then
    Dispose(KeyMap);
end;

end.