{Tera Term
 Copyright(C) 1994-1996 T. Teranishi
 All rights reserved.}

{Misc. routines}
unit TTLib;

interface

uses WinTypes, Strings, WinDos;

procedure StrUpperJ(Str: PChar);
function GetFileNamePosJ(PathName: PChar; var DirLen,FNPos: integer): boolean;
function ExtractFileNameJ(PathName,FileName: PChar): boolean;
function ExtractDirNameJ(PathName,DirName: PChar): boolean;
procedure FitFileName(FileName, DefExt: PChar);
procedure Str2Hex(Str, Hex: PChar; Len, MaxHexLen: integer; ConvSP: bool);
function Hex2Str(Hex, Str: PChar; MaxStrLen: integer): integer;

implementation

procedure StrUpperJ(Str: PChar);
var
  i, L: integer;
  b: byte;
begin
  if Str=nil then exit;
  L := StrLen(Str);
  i := 0;
  while i<L do
  begin
    b := byte(Str[i]);
    inc(i);
    case b of
      Ord('a')..Ord('z'): {lower -> upper}
        Str[i-1] := char(b - $20); 
      $81..$9F,
      $E0..$FC: inc(i); {SJIS first byte}
    end;
  end;
end;

function GetFileNamePosJ(PathName: PChar; var DirLen,FNPos: integer): boolean;
var
  i, j, k, L: integer;
  b: byte;
begin
  GetFileNamePosJ := false;
  DirLen := 0;
  FNPos := 0;
  if PathName=nil then exit;
  L := StrLen(PathName);
  i := 0;
  if (L>=2) and (PathName[1]=':') then
    i := 2;
  if PathName[i]='\' then inc(i);
  j := i;
  k := i;
  while i<L do
  begin
    b := byte(PathName[i]);
    inc(i);
    case b of
      ord('*'),
      ord(':'): exit;
      ord('\'): begin
          j := i - 1;
          k := i;
        end;
      $81..$9F,
      $E0..$FC: inc(i); {SJIS first byte}
    end;
  end;
  DirLen := j;
  FNPos := k;
  GetFileNamePosJ := true;
end;

function ExtractFileNameJ(PathName,FileName: PChar): boolean;
var
  i, j: integer;
begin
  ExtractFileNameJ := false;
  if FileName=nil then exit;
  if not GetFileNamePosJ(PathName,i,j) then exit;
  StrLCopy(FileName,@PathName[j],fsFileName+fsExtension);
  ExtractFileNameJ := StrLen(FileName)>0;
end;

function ExtractDirNameJ(PathName,DirName: PChar): boolean;
var
  i, j: integer;
begin
  ExtractDirNameJ := false;
  if DirName=nil then exit;
  if not GetFileNamePosJ(PathName,i,j) then exit;
  ExtractDirNameJ := true;
  if i > fsDirectory then i := fsDirectory;
  Move(PathName[0],DirName[0],i);
  DirName[i] := #0;
end;

procedure FitFileName(FileName, DefExt: PChar);
var
  i, j, k: integer;
  Temp: array[0..fsFileName+fsExtension] of char;
  b: byte;
begin
  i := 0;
  j := 0;
  repeat
    b := byte(FileName[i]);
    inc(i);
    if (b<>0) and (b<>ord('.')) and
      (j<fsFileName) then
    begin
      Temp[j] := char(b);
      inc(j);
    end;
  until (b=ord('.')) or (b=0);
  if b=0 then
  begin
    Temp[j] := #0;
    if DefExt<> nil then
      StrCat(Temp,DefExt);
  end
  else begin
    Temp[j] := '.';
    inc(j);
    k := 0;
    repeat
      b := byte(FileName[i]);
      inc(i);
      if (b<>0) and (b<>ord('.')) and 
        (k<fsExtension-1) then
      begin
        Temp[j] := char(b);
        inc(j);
        inc(k);
      end;
    until (b=ord('.')) or (b=0);
    Temp[j] := #0;
  end;
  StrLCopy(FileName,Temp,fsFileName+fsExtension);
end;

procedure Str2Hex(Str, Hex: PChar; Len, MaxHexLen: integer; ConvSP: bool);
var
  b, low: byte;
  i, j: integer;
begin
  if ConvSP then
    low := $20
  else
    low := $1F;

  j := 0;
  for i := 0 to Len-1 do
  begin
    b := byte(Str[i]);
    if (b<>Ord('$')) and (b>Low) and (b<$7f) then
    begin
      if j < MaxHexLen then
      begin
        Hex[j] := char(b);
        inc(j);
      end;
    end
    else begin
      if j < MaxHexLen-2 then
      begin
        Hex[j] := '$';
        inc(j);
        case b of
          $00..$9F: Hex[j] := char(b shr 4 + $30);
          $A0..$FF: Hex[j] := char(b shr 4 + $37);
        end;
        inc(j);
        case b and $0F of
          $0..$9: Hex[j] := char(b and $0F + $30);
          $A..$F: Hex[j] := char(b and $0F + $37);
        end;
        inc(j);
      end;
    end;
  end;
  Hex[j] := #0;
end;

  function ConvHexChar(b: byte): byte;
  begin
    case b of
      ord('0')..ord('9'):
        ConvHexChar := b - $30;
      ord('A')..ord('F'):
        ConvHexChar := b - $37;
      ord('a')..ord('f'):
        ConvHexChar := b - $57;
    else
      ConvHexChar := 0;
    end;
  end;

function Hex2Str(Hex, Str: PChar; MaxStrLen: integer): integer;
var
  b, c: byte;
  i, imax, j: integer;
begin
  j := 0;
  imax := StrLen(Hex);
  i := 0;
  while (i < imax) and (j<MaxStrLen)do
  begin
    b := byte(Hex[i]);
    if b=ord('$') then
    begin
      inc(i);
      if i < imax then
        c := byte(Hex[i])
      else
        c := $30;
      b := ConvHexChar(c) shl 4;
      inc(i);
      if i < imax then
        c := byte(Hex[i])
      else
        c := $30;
      b := b + ConvHexChar(c);
    end;

    Str[j] := char(b);
    inc(j);

    inc(i);
  end;
  Hex2Str := j;
end;

end.