{Tera Term
 Copyright(C) 1994-1996 T. Teranishi
 All rights reserved.}

{TTMACRO.EXE, TTL parser}
unit TTMParse;

interface
{$I teraterm.inc}

{$IFDEF Delphi}
uses WinTypes, OWindows, WinProcs, Strings, TTMDlg;
{$ELSE}
uses WinTypes, WObjects, WinProcs, Strings, TTMDlg;
{$ENDIF}

const
  IdTTLRun  = 1;
  IdTTLWait = 2;
  IdTTLWait2 = 3;
  IdTTLInitDDE = 4;
  IdTTLPause = 5;
  IdTTLWaitCmndEnd = 6;
  IdTTLWaitConnect = 7;
  IdTTLEnd  = 8;

  ErrCloseParent = 1;
  ErrCantCall = 2;
  ErrCantConnect = 3;
  ErrCantOpen = 4;
  ErrDivByZero = 5;
  ErrInvalidCtl = 6;
  ErrLabelAlreadyDef = 7;
  ErrLabelReq = 8;
  ErrLinkFirst = 9;
  ErrStackOver = 10;
  ErrSyntax = 11;
  ErrTooManyLabels = 12;
  ErrTooManyVar = 13;
  ErrTypeMismatch = 14;
  ErrVarNotInit = 15;

  TypUnknown = 0;
  TypInteger = 1;
  TypLogical = 2;
  TypString = 3;
  TypLabel = 4;

  RsvBeep       = 1;
  RsvBPlusRecv  = 2;
  RsvBPlusSend  = 3;
  RsvCall       = 4;
  RsvChangeDir  = 5;
  RsvCloseTT    = 6;
  RsvConnect    = 7;
  RsvElse       = 8;
  RsvElseIf     = 9;
  RsvEnd        = 10;
  RsvEndIf      = 11;
  RsvEndWhile   = 12;
  RsvExec       = 13;
  RsvExecCmnd   = 14;
  RsvExit       = 15;
  RsvFileClose  = 16;
  RsvFileConcat = 17;
  RsvFileCopy   = 18;
  RsvFileCreate = 19;
  RsvFileDelete = 20;
  RsvFileOpen   = 21;
  RsvFileReadln = 22;
  RsvFileRename = 23;
  RsvFileSearch = 24;
  RsvFileSeek   = 25;
  RsvFileStrSeek = 26;
  RsvFileWrite  = 27;
  RsvFileWriteln = 28;
  RsvFor        = 29;
  RsvGetDate    = 30;
  RsvGetTime    = 31;
  RsvGoto       = 32;
  RsvIf         = 33;
  RsvInclude    = 34;
  RsvInputBox   = 35;
  RsvInt2Str    = 36;
  RsvKmtRecv    = 37;
  RsvKmtSend    = 38;
  RsvLogClose   = 39;
  RsvLogOpen    = 40;
  RsvLogPause   = 41;
  RsvLogStart   = 42;
  RsvLogWrite   = 43;
  RsvMessageBox = 44;
  RsvNext       = 45;
  RsvPasswordBox = 46;
  RsvPause      = 47;
  RsvQuickVANRecv = 48;
  RsvQuickVANSend = 49;
  RsvReturn     = 50;
  RsvSend       = 51;
  RsvSendFile   = 52;
  RsvSendln     = 53;
  RsvShow       = 54;
  RsvShowTT     = 55;
  RsvStr2Int    = 56;
  RsvStrCompare = 57;
  RsvStrConcat  = 58;
  RsvStrCopy    = 59;
  RsvStrLen     = 60;
  RsvStrScan    = 61;
  RsvThen       = 62;
  RsvWait       = 63;
  RsvWaitRecv   = 64;
  RsvWhile      = 65;
  RsvXmodemRecv = 66;
  RsvXmodemSend = 67;
  RsvYesNoBox   = 68;
  RsvZmodemRecv = 69;
  RsvZmodemSend = 70;

  RsvNot        = 101;
  RsvAnd        = 102;
  RsvOr         = 103;
  RsvXor        = 104;
  RsvMul        = 105;
  RsvPlus       = 106;
  RsvMinus      = 107;
  RsvDiv        = 108;
  RsvLT         = 109;
  RsvEQ         = 110;
  RsvGT         = 111;
  RsvLE         = 112;
  RsvNE         = 113;
  RsvGE         = 114;

const
  MaxNameLen = 32;
  MaxStrLen = 256;

type
  PName = ^TName;
  TName = array[0..MaxNameLen] of char;
  PStrVal = ^TStrVal;
  TStrVal = array[0..MaxStrLen] of char;

var
  LineBuff: TStrVal;
  LinePtr: word;
  LineLen: word;
  TTLStatus: word;

const
  MaxNumOfIntVar = 128;
  MaxNumOfStrVar = 128;
  MaxNumOfLabVar = 256;
  IntVarIdOff = 0;
  StrVarIdOff = IntVarIdOff + MaxNumOfIntVar;
  LabVarIdOff = StrVarIdOff + MaxNumOfStrVar;
  MaxNumOfName = MaxNumOfIntVar + MaxNumOfStrVar + MaxNumOfLabVar;
  NameBuffLen = MaxNumOfName*MaxNameLen;
  StrBuffLen = MaxNumOfStrVar*MaxStrLen;

type
  PNameBuff = ^TNameBuff;
  TNameBuff = array[0..NameBuffLen-1] of char;
  PStrBuff = ^TStrBuff;
  TStrBuff = array[0..StrBuffLen-1] of char;
var
  HNameBuff: THandle;
  NameBuff: PNameBuff;
  HStrBuff: THandle;
  StrBuff: PNameBuff;
  IntVal:  array[0..MaxNumOfIntVar-1] of integer;
  LabVal:  array[0..MaxNumOfLabVar-1] of word;
  LabLevel: array[0..MaxNumOfLabVar-1] of byte;
  IntVarCount, StrVarCount, LabVarCount: word;

procedure InitVar;
procedure EndVar;
procedure LockVar;
procedure UnlockVar;
procedure DispErr(Err: word);
function GetFirstChar: char;
function GetIdentifier(Name: PChar): BOOL;
function GetReservedWord(var WordId: word): BOOL;
function GetLabelName(Name: PChar): BOOL;
function GetString(Str: PChar; var Err: word): BOOL;
function CheckVar(Name: PChar; var VarType, VarId: word): BOOL;
function NewIntVar(Name: PChar; InitVal: integer): BOOL;
function NewStrVar(Name: PChar; InitVal: PChar): BOOL;
function NewLabVar(Name: PChar; InitVal,ILevel: word): BOOL;
procedure DelLabVar(ILevel: word);
function GetExpression(var ValType: word; var Val: integer; var Err: word): BOOL;
procedure GetIntVal(var Val: integer; var Err: word);
procedure GetIntVar(var VarId, Err: word);
procedure GetStrVal(Str: PChar; var Err: word);
procedure GetStrVar(var VarId, Err: word);
procedure SetStrVal(VarId: word; Str: PChar);
function StrVarPtr(VarId: word): PChar;

implementation

procedure InitVar;
begin
  HNameBuff := GlobalAlloc(GMEM_MOVEABLE or GMEM_ZEROINIT,NameBuffLen);
  if HNameBuff=0 then PostQuitMessage(0);
  NameBuff := nil; 

  HStrBuff := GlobalAlloc(GMEM_MOVEABLE or GMEM_ZEROINIT,StrBuffLen);
  if HStrBuff=0 then PostQuitMessage(0);
  StrBuff := nil;

  IntVarCount := 0;
  LabVarCount := 0;
  StrVarCount := 0;
end;

procedure EndVar;
begin
  UnlockVar;
  GlobalFree(HNameBuff);
  GlobalFree(HStrBuff);
end;

procedure LockVar;
begin
  if NameBuff=nil then
    NameBuff := GlobalLock(HNameBuff);
  if NameBuff=nil then
    PostQuitMessage(0);

  if StrBuff=nil then
    StrBuff := GlobalLock(HStrBuff);
  if StrBuff=nil then
    PostQuitMessage(0);
end;

procedure UnlockVar;
begin
  if NameBuff<>nil then
    GlobalUnlock(HNameBuff);
  NameBuff := nil;

  if StrBuff<>nil then
    GlobalUnlock(HStrBuff);
  StrBuff := nil;
end;

procedure DispErr(Err: word);
var
  Msg: array[0..40] of char;
  i: integer;
  b: char;
begin
  case Err of
    ErrCloseParent: StrCopy(Msg,'")" expected.');
    ErrCantCall: StrCopy(Msg,'Can''t call sub.');
    ErrCantConnect: StrCopy(Msg,'Can''t link macro.');
    ErrCantOpen: StrCopy(Msg,'Can''t open file.');
    ErrDivByZero: StrCopy(Msg,'Divide by zero.');
    ErrInvalidCtl: StrCopy(Msg,'Invalid control.');
    ErrLabelAlreadyDef: StrCopy(Msg,'Label already defined.');
    ErrLabelReq: StrCopy(Msg,'Label requiered.');
    ErrLinkFirst: StrCopy(Msg,'Link macro first.');
    ErrStackOver: StrCopy(Msg,'Stack overflow.');
    ErrSyntax: StrCopy(Msg,'Syntax error.');
    ErrTooManyLabels: StrCopy(Msg,'Too many labels.');
    ErrTooManyVar: StrCopy(Msg,'Too many variables.');
    ErrTypeMismatch: StrCopy(Msg,'Type mismatch.');
    ErrVarNotInit: StrCopy(Msg,'Variable not initialized.');
  end;

  i := Application^.ExecDialog(
         New(PErrDlg,Init(Msg,LineBuff)) );
  if i=1 then TTLStatus := IdTTLEnd;
end;

function CheckReservedWord(Str: PChar; var WordId: word): BOOL;
begin
  WordId := 0;

  if StrIComp(Str,'beep')=0 then WordId := RsvBeep
  else if StrIComp(Str,'bplusrecv')=0 then WordId := RsvBPlusRecv
  else if StrIComp(Str,'bplussend')=0 then WordId := RsvBPlusSend
  else if StrIComp(Str,'call')=0 then WordId := RsvCall
  else if StrIComp(Str,'changedir')=0 then WordId := RsvChangeDir
  else if StrIComp(Str,'closett')=0 then WordId := RsvCloseTT
  else if StrIComp(Str,'connect')=0 then WordId := RsvConnect
  else if StrIComp(Str,'else')=0 then WordId := RsvElse
  else if StrIComp(Str,'elseif')=0 then WordId := RsvElseIf
  else if StrIComp(Str,'end')=0 then WordId := RsvEnd
  else if StrIComp(Str,'endif')=0 then WordId := RsvEndIf
  else if StrIComp(Str,'endwhile')=0 then WordId := RsvEndWhile
  else if StrIComp(Str,'exec')=0 then WordId := RsvExec
  else if StrIComp(Str,'execcmnd')=0 then WordId := RsvExecCmnd
  else if StrIComp(Str,'exit')=0 then WordId := RsvExit
  else if StrIComp(Str,'fileclose')=0 then WordId := RsvFileClose
  else if StrIComp(Str,'fileconcat')=0 then WordId := RsvFileConcat
  else if StrIComp(Str,'filecopy')=0 then WordId := RsvFileCopy
  else if StrIComp(Str,'filecreate')=0 then WordId := RsvFileCreate
  else if StrIComp(Str,'filedelete')=0 then WordId := RsvFileDelete
  else if StrIComp(Str,'fileopen')=0 then WordId := RsvFileOpen
  else if StrIComp(Str,'filereadln')=0 then WordId := RsvFileReadln
  else if StrIComp(Str,'filerename')=0 then WordId := RsvFileRename
  else if StrIComp(Str,'filesearch')=0 then WordId := RsvFileSearch
  else if StrIComp(Str,'fileseek')=0 then WordId := RsvFileSeek
  else if StrIComp(Str,'filestrseek')=0 then WordId := RsvFileStrSeek
  else if StrIComp(Str,'filewrite')=0 then WordId := RsvFileWrite
  else if StrIComp(Str,'filewriteln')=0 then WordId := RsvFileWriteln
  else if StrIComp(Str,'for')=0 then WordId := RsvFor
  else if StrIComp(Str,'getdate')=0 then WordId := RsvGetDate
  else if StrIComp(Str,'gettime')=0 then WordId := RsvGetTime
  else if StrIComp(Str,'goto')=0 then WordId := RsvGoto
  else if StrIComp(Str,'if')=0 then WordId := RsvIf
  else if StrIComp(Str,'include')=0 then WordId := RsvInclude 
  else if StrIComp(Str,'inputbox')=0 then WordId := RsvInputBox
  else if StrIComp(Str,'int2str')=0 then WordId := RsvInt2Str
  else if StrIComp(Str,'kmtrecv')=0 then WordId := RsvKmtRecv
  else if StrIComp(Str,'kmtsend')=0 then WordId := RsvKmtSend
  else if StrIComp(Str,'logclose')=0 then WordId := RsvLogClose
  else if StrIComp(Str,'logopen')=0 then WordId := RsvLogOpen
  else if StrIComp(Str,'logpause')=0 then WordId := RsvLogPause
  else if StrIComp(Str,'logstart')=0 then WordId := RsvLogStart
  else if StrIComp(Str,'logwrite')=0 then WordId := RsvLogWrite
  else if StrIComp(Str,'messagebox')=0 then WordId := RsvMessageBox
  else if StrIComp(Str,'next')=0 then WordId := RsvNext
  else if StrIComp(Str,'passwordbox')=0 then WordId := RsvPasswordBox
  else if StrIComp(Str,'pause')=0 then WordId := RsvPause
  else if StrIComp(Str,'quickvanrecv')=0 then WordId := RsvQuickVANRecv
  else if StrIComp(Str,'quickvansend')=0 then WordId := RsvQuickVANSend
  else if StrIComp(Str,'return')=0 then WordId := RsvReturn
  else if StrIComp(Str,'send')=0 then WordId := RsvSend
  else if StrIComp(Str,'sendfile')=0 then WordId := RsvSendFile
  else if StrIComp(Str,'sendln')=0 then WordId := RsvSendln
  else if StrIComp(Str,'show')=0 then WordId := RsvShow
  else if StrIComp(Str,'showtt')=0 then WordId := RsvShowTT
  else if StrIComp(Str,'str2int')=0 then WordId := RsvStr2Int
  else if StrIComp(Str,'strcompare')=0 then WordId := RsvStrCompare
  else if StrIComp(Str,'strconcat')=0 then WordId := RsvStrConcat
  else if StrIComp(Str,'strcopy')=0 then WordId := RsvStrCopy
  else if StrIComp(Str,'strlen')=0 then WordId := RsvStrLen
  else if StrIComp(Str,'strscan')=0 then WordId := RsvStrScan
  else if StrIComp(Str,'then')=0 then WordId := RsvThen
  else if StrIComp(Str,'wait')=0 then WordId := RsvWait
  else if StrIComp(Str,'waitrecv')=0 then WordId := RsvWaitRecv
  else if StrIComp(Str,'while')=0 then WordId := RsvWhile
  else if StrIComp(Str,'xmodemrecv')=0 then WordId := RsvXmodemRecv
  else if StrIComp(Str,'xmodemsend')=0 then WordId := RsvXmodemSend
  else if StrIComp(Str,'yesnobox')=0 then WordId := RsvYesNoBox
  else if StrIComp(Str,'zmodemrecv')=0 then WordId := RsvZmodemRecv
  else if StrIComp(Str,'zmodemsend')=0 then WordId := RsvZmodemSend

  else if StrIComp(Str,'not')=0 then WordId := RsvNot
  else if StrIComp(Str,'and')=0 then WordId := RsvAnd
  else if StrIComp(Str,'or')=0 then WordId := RsvOr
  else if StrIComp(Str,'xor')=0 then WordId := RsvXor;

  CheckReservedWord := WordId<>0;
end;

function GetFirstChar: char;
var
  b: char;
begin
  GetFirstChar := #0;
  if LinePtr<LineLen then
    b := LineBuff[LinePtr]
  else exit;

  while (LinePtr<LineLen) and ((b=#$20) or (b=#$09)) do
  begin
    inc(LinePtr);
    if LinePtr<LineLen then b := LineBuff[LinePtr];
  end;
  if b>#$20 then
  begin
    GetFirstChar := b;
    inc(LinePtr);
  end;
end;

function GetIdentifier(Name: PChar): BOOL;
var
  i: integer;
  b: char;
begin
  GetIdentifier := false;

  FillChar(Name[0],MaxNameLen+1,0);

  b := GetFirstChar;
  if b=#0 then exit;

  {Check first character of identifier}
  if ((b<'A') or (b>'Z')) and
     (b<>'_') and
     ((b<'a') or (b>'z')) then
  begin
    dec(LinePtr);
    exit;
  end;

  GetIdentifier := true;
  Name[0] := b;
  i := 1;

  if LinePtr<LineLen then b := LineBuff[LinePtr];
  while (LinePtr<LineLen) and
        ( (b>='0') and (b<='9') or
          (b>='A') and (b<='Z') or
          (b>='_') or
          (b>='a') and (b<='z') ) do
  begin
    if i<MaxNameLen then
    begin
      Name[i] := b;
      inc(i);
    end;
    inc(LinePtr);
    if LinePtr<LineLen then b := LineBuff[LinePtr];
  end;

end;

function GetReservedWord(var WordId: word): BOOL;
var
  Name: TName;
  P: word;
begin
  GetReservedWord := false;
  P := LinePtr;
  if not GetIdentifier(Name) then exit;
  if not CheckReservedWord(Name,WordId) then
  begin
    LinePtr := P;
    exit;
  end;
  if 0<WordId then
    GetReservedWord := true
  else
    LinePtr := P;
end;

function GetOperator(var WordId: word): BOOL;
var
  Name: TName;
  P: word;
  b: Char;
begin
  GetOperator := false;
  P := LinePtr;
  b := GetFirstChar;
  case b of
    #0: exit;
    '*': WordId := RsvMul;
    '+': WordId := RsvPlus;
    '-': WordId := RsvMinus;
    '/': WordId := RsvDiv;
    '<': WordId := RsvLT;
    '=': WordId := RsvEQ;
    '>': WordId := RsvGT;
  else
    begin
      dec(LinePtr);
      if not GetReservedWord(WordId) or (WordId<100) then 
      begin
        LinePtr := P;
        exit;
      end;
    end;
  end;

  if ((WordId=RsvLT) or (WordId=RsvGT)) and
     (LinePtr<LineLen) then
  begin
    b := LineBuff[LinePtr];
    if b='=' then
    begin
      if WordId=RsvLT then
        WordId:=RsvLE
      else
        WordId:=RsvGE;
      inc(LinePtr);
    end
    else if (b='>') and (WordId=RsvLT) then
    begin
      WordId := RsvNE;
      inc(LinePtr);
    end;
  end;
  GetOperator := true;
end;

function GetLabelName(Name: PChar): BOOL;
var
  i: integer;
  b: char;
begin
  GetLabelName := false;

  FillChar(Name[0],MaxNameLen+1,0);

  b := GetFirstChar;
  if b=#0 then exit;
  Name[0] := b;

  i := 1;
  if LinePtr<LineLen then b := LineBuff[LinePtr];
  while (LinePtr<LineLen) and
        ( (b>='0') and (b<='9') or
          (b>='A') and (b<='Z') or
          (b>='_') or
          (b>='a') and (b<='z') ) do
  begin
    if i<MaxNameLen then
    begin
      Name[i] := b;
      inc(i);
    end;
    inc(LinePtr);
    if LinePtr<LineLen then b := LineBuff[LinePtr];
  end;

  GetLabelName := StrLen(Name)>0; 
end;

function GetString(Str: PChar; var Err: word): BOOL;

  function GetQuotedStr(q: char; var i: word): integer;
  var
    b: char;
  begin
    GetQuotedStr := 0;
    b:=#0;
    if LinePtr<LineLen then b := LineBuff[LinePtr];
    while (LinePtr<LineLen) and (b>=' ') and (b<>q) do
    begin
      if i<MaxStrLen then
      begin
        Str[i] := b;
        inc(i);
      end;

      inc(LinePtr);
      if LinePtr<LineLen then b := LineBuff[LinePtr];
    end;
    if b=q then
    begin
      if LinePtr<LineLen then inc(LinePtr)
    end
    else GetQuotedStr := ErrSyntax;
  end;

  function GetCharByCode(var i: word): word;
  var
    b: char;
    n: word;
  begin
    GetCharByCode := ErrSyntax;
    b:=#0;
    n := 0;
    if LinePtr<LineLen then b := LineBuff[LinePtr];
    if (b<'0') or (b>'9') then exit;

    while (LinePtr<LineLen) and (b>='0') and (b<='9') do
    begin
      n := n * 10 + byte(b) - $30;
      inc(LinePtr);
      if LinePtr<LineLen then b := LineBuff[LinePtr];
    end;

    if (n<=0) or (n>255) then exit;
    GetCharByCode := 0;

    if i<MaxStrLen then
    begin
      Str[i] := char(n);
      inc(i);
    end;
  end;

var
  q: char;
  i: word;
begin
  GetString := false;
  Err := 0;
  FillChar(Str[0],MaxStrLen+1,0);

  q := GetFirstChar;
  if q=#0 then exit;
  dec(LinePtr);
  if (q=#$22) or (q=#$27) or (q='#') then
    GetString := true
  else
    exit;

  i := 0;
  while ((q=#$22) or (q=#$27) or (q='#')) and (Err=0) do
  begin
    inc(LinePtr);
    case q of
      #$22,#$27: Err := GetQuotedStr(q,i);
      '#': Err := GetCharByCode(i);
    end;
    q := LineBuff[LinePtr];
  end;

end;

function GetNumber(var Num: integer): BOOL;
var
  b: char;
  i: integer;
begin
  GetNumber := false;
  Num := 0;

  b := GetFirstChar;
  if b=#0 then exit;
  if (b>='0') and (b<='9') then
    Num := byte(b)-$30
  else begin
    dec(LinePtr);
    exit;
  end;
  GetNumber := true;

  if LinePtr<LineLen then b := LineBuff[LinePtr];
  while (LinePtr<LineLen) and (b>='0') and (b<='9') do
  begin
    Num := Num*10 - $30 + byte(b);
    inc(LinePtr);
    if LinePtr<LineLen then b := LineBuff[LinePtr];
  end;
end;

function CheckVar(Name: PChar; var VarType, VarId: word): BOOL;
var
  i, P: word;
begin
  CheckVar := true;
  VarType := TypUnknown;

  i := 0;
  while i<IntVarCount do
  begin
    P := (i+IntVarIdOff)*MaxNameLen;
    if StrIComp(@NameBuff^[P],Name)=0 then
    begin
      VarType := TypInteger;
      VarId := i;
      exit;
    end;
    inc(i);
  end;

  i := 0;
  while i<StrVarCount do
  begin
    P := (i+StrVarIdOff)*MaxNameLen;
    if StrIComp(@NameBuff^[P],Name)=0 then
    begin
      VarType := TypString;
      VarId := i;
      exit;
    end;
    inc(i);
  end;

  i := 0;
  while i<LabVarCount do
  begin
    P := (i+LabVarIdOff)*MaxNameLen;
    if StrIComp(@NameBuff^[P],Name)=0 then
    begin
      VarType := TypLabel;
      VarId := i;
      exit;
    end;
    inc(i);
  end;

  CheckVar := false;
end;

function NewIntVar(Name: PChar; InitVal: integer): BOOL;
var
  P: word;
begin
  NewIntVar := false;
  if IntVarCount>=MaxNumOfIntVar then exit;
  P := (IntVarIdOff+IntVarCount)*MaxNameLen;
  StrCopy(@NameBuff^[P],Name);
  IntVal[IntVarCount] := InitVal;
  inc(IntVarCount);
  NewIntVar := true;
end;

function NewStrVar(Name: PChar; InitVal: PChar): BOOL;
var
  P: word;
begin
  NewStrVar := false;
  if StrVarCount>=MaxNumOfStrVar then exit;
  P := (StrVarIdOff+StrVarCount)*MaxNameLen;
  StrCopy(@NameBuff^[P],Name);
  P := StrVarCount*MaxStrLen;
  StrCopy(@StrBuff^[P],InitVal);
  inc(StrVarCount);
  NewStrVar := true;
end;

function NewLabVar(Name: PChar; InitVal, ILevel: word): BOOL;
var
  P: word;
begin
  NewLabVar := false;
  if LabVarCount>=MaxNumOfLabVar then exit;

  P := (LabVarIdOff+LabVarCount)*MaxNameLen;
  StrCopy(@NameBuff^[P],Name);
  LabVal[LabVarCount] := InitVal;
  LabLevel[LabVarCount] := byte(ILevel);
  inc(LabVarCount);
  NewLabVar := true;
end;

procedure DelLabVar(ILevel: word);
begin
  while (LabVarCount>0) and (LabLevel[LabVarCount-1]>=ILevel) do
    dec(LabVarCount);
end;

function GetFactor(var ValType: word; var Val: integer; var Err: word): BOOL;
var
  Name: TName;
  P, Id, Er: word;
begin
  GetFactor := true;
  P := LinePtr;
  Err := 0;
  if GetIdentifier(Name) then
  begin
    if CheckReservedWord(Name,Id) then
    begin
      if Id=RsvNot then
      begin
        if GetFactor(ValType,Val,Err) then
        begin
          if (Err=0) and (ValType<>TypInteger) then
            Err := ErrTypeMismatch;
          Val := not Val;
        end
        else
          Err := ErrSyntax;
      end
      else
        Err := ErrSyntax;
    end
    else if CheckVar(Name, ValType, Id) then
    begin
      case ValType of
        TypInteger: Val := IntVal[Id];
        TypString: Val := Id;
      end;
    end
    else
      Err := ErrVarNotInit;
  end
  else if GetNumber(Val) then
    ValType := TypInteger
  else if GetOperator(Id) then
  begin
    if (Id=RsvPlus) or (Id=RsvMinus) then
    begin
      if GetFactor(ValType,Val,Err) then
      begin
        if (Err=0) and (ValType<>TypInteger) then
          Err := ErrTypeMismatch;
        if Id=RsvMinus then Val := -Val;
      end
      else
        Err := ErrSyntax;
    end
    else
      Err := ErrSyntax;
  end
  else if GetFirstChar='(' then
  begin
    if GetExpression(ValType,Val,Err) then
    begin
      if (Err=0) and (GetFirstChar<>')') then
        Err := ErrCloseParent;
    end
    else
      Err := ErrSyntax;
  end
  else begin
    GetFactor := false;
    Err := 0
  end;

  if Err<>0 then LinePtr := P;
end;

function GetTerm(var ValType: word; var Val: integer; var Err: word): BOOL;
var
  P1, P2, Type1, Type2, Er: word;
  Val1, Val2: integer;
  WId: word;
begin
  GetTerm := false;
  P1 := LinePtr;
  if not GetFactor(Type1,Val1,Er) then exit;
  GetTerm := true;
  ValType := Type1;
  Val := Val1;
  Err := Er;
  if Er<>0 then
  begin
    LinePtr := P1;
    exit;
  end;
  if Type1<>TypInteger then exit;

  repeat
    P2 := LinePtr;
    if not GetOperator(WId) then exit;

    case WId of 
      RsvAnd:;
      RsvMul:;
      RsvDiv:;
    else
      begin
        LinePtr := P2;
        exit;
      end;
    end;

    if not GetFactor(Type2,Val2,Er) then
    begin
      Err := ErrSyntax;
      LinePtr := P1;
      exit;
    end;

    if Er<>0 then
    begin
      Err := Er;
      LinePtr := P1;
      exit;
    end;

    if Type2<>TypInteger then
    begin
      Err := ErrTypeMismatch;
      LinePtr := P1;
      exit;
    end;

    case WId of
      RsvAnd: Val1 := Val1 and Val2;
      RsvMul: Val1 := Val1  *  Val2;
      RsvDiv: if Val2<>0 then
                Val1 := Val1 div Val2
              else begin
                Err := ErrDivByZero;
                LinePtr := P1;
                exit;
              end;
    end;

    Val := Val1;
  until false;
end;

function GetSimpleExpression(var ValType: word; var Val: integer; var Err: word): BOOL;
var
  P1, P2, Type1,Type2, Er: word;
  Val1,Val2: integer;
  WId: word;
begin
  GetSimpleExpression := false;
  P1 := LinePtr;
  if not GetTerm(Type1,Val1,Er) then exit;
  GetSimpleExpression := true;
  ValType := Type1;
  Val := Val1;
  Err := Er;
  if Er<>0 then
  begin
    LinePtr := P1;
    exit;
  end;
  if Type1<>TypInteger then exit;

  repeat
    P2 := LinePtr;
    if not GetOperator(WId) then exit;

    case WId of 
      RsvOr:;
      RsvXor:;
      RsvPlus:;
      RsvMinus:;
    else
      begin
        LinePtr := P2;
        exit;
      end;
    end;

    if not GetTerm(Type2,Val2,Er) then
    begin
      Err := ErrSyntax;
      LinePtr := P1;
      exit;
    end;

    if Er<>0 then
    begin
      Err := Er;
      LinePtr := P1;
      exit;
    end;

    if Type2<>TypInteger then
    begin
      Err := ErrTypeMismatch;
      LinePtr := P1;
      exit;
    end;

    case WId of
      RsvOr:    Val1 := Val1 or  Val2;
      RsvXor:   Val1 := Val1 xor Val2;
      RsvPlus:  Val1 := Val1  +  Val2;
      RsvMinus: Val1 := Val1  -  Val2;
    end;
    Val := Val1;
  until false;
end;

function GetExpression(var ValType: word; var Val: integer; var Err: word): BOOL;
var
  P1,P2, Type1,Type2, Er: word;
  Val1, Val2: integer;
  WId: word;
begin
  P1 := LinePtr;
  if not GetSimpleExpression(Type1,Val1,Er) then
  begin
    GetExpression := false;
    LinePtr := P1;
    exit;
  end;
  GetExpression := true;
  ValType := Type1;
  Val := Val1;
  Err := Er;
  if Er<>0 then
  begin
    LinePtr := P1;
    exit;
  end;
  if Type1<>TypInteger then exit;

  P2 := LinePtr;
  if not GetOperator(WId) then exit;

  case WId of 
    RsvLT:;
    RsvEQ:;
    RsvGT:;
    RsvLE:;
    RsvNE:;
    RsvGE:;
  else
    begin
      LinePtr := P2;
      exit;
    end;
  end;

  if not GetSimpleExpression(Type2,Val2,Er) then
  begin
    Err := ErrSyntax;
    LinePtr := P1;
    exit;
  end;

  if Er<>0 then
  begin
    Err := Er;
    LinePtr := P1;
    exit;
  end;

  if Type2<>TypInteger then
  begin
    Err := ErrTypeMismatch;
    exit;
  end;

  Val := 0;
  case WId of
    RsvLT: if Val1<Val2 then Val := 1;
    RsvEQ: if Val1=Val2 then Val := 1;
    RsvGT: if Val1>Val2 then Val := 1;
    RsvLE: if Val1<=Val2 then Val := 1;
    RsvNE: if Val1<>Val2 then Val := 1;
    RsvGE: if Val1>=Val2 then Val := 1;
  end;
end;

procedure GetIntVal(var Val: integer; var Err: word);
var
  ValType: word;
begin
  if Err<>0 then exit;
  if not GetExpression(ValType,Val,Err) then
  begin
    Err := ErrSyntax;
    exit;
  end;
  if Err<>0 then exit;
  if ValType<>TypInteger then
    Err := ErrTypeMismatch;
end;

procedure GetIntVar(var VarId, Err: word);
var
  Name: TName;
  VarType: word;
begin
  if Err<>0 then exit;

  if GetIdentifier(Name) then
  begin
    if CheckVar(Name,VarType,VarId) then
    begin
      if VarType<>TypInteger then
        Err := ErrTypeMismatch;
    end
    else begin
      if NewIntVar(Name,0) then
        CheckVar(Name,VarType,VarId)
      else
        Err := ErrTooManyVar;
    end;
  end
  else
    Err := ErrSyntax;
end;

procedure GetStrVal(Str: PChar; var Err: word);
var
  VarType, VarId: word;
begin
  Str[0] := #0;
  if Err<>0 then exit;

  if GetString(Str,Err) then
    exit
  else if GetExpression(VarType,integer(VarId),Err) then
  begin
    if Err<>0 then exit;
    if VarType<>TypString then
      Err := ErrTypeMismatch
    else
      StrCopy(Str,@StrBuff^[VarId*MaxStrLen]);
  end
  else
    Err := ErrSyntax;
end;

procedure GetStrVar(var VarId, Err: word);
var
  Name: TName;
  VarType: word;
begin
  if Err<>0 then exit;

  if GetIdentifier(Name) then
  begin
    if CheckVar(Name,VarType,VarId) then
    begin
      if VarType<>TypString then
        Err := ErrTypeMismatch;
    end
    else begin
      if NewStrVar(Name,#0) then
        CheckVar(Name,VarType,VarId)
      else
        Err := ErrTooManyVar;
    end;
  end
  else
    Err := ErrSyntax;
end;

procedure SetStrVal(VarId: word; Str: PChar);
begin
  StrCopy(@StrBuff^[VarId*MaxStrLen],Str);
end;

function StrVarPtr(VarId: word): PChar;
begin
  StrVarPtr := @StrBuff^[VarId*MaxStrLen];
end;

end.