{Tera Term
 Copyright(C) 1994-1996 T. Teranishi
 All rights reserved.}

{TTTEK.DLL, for TEK window}
library TTTEK;
{$R tttek.res}
{$I teraterm.inc}

{$IFDEF Delphi}
uses WinTypes, WinProcs, OWindows, Strings, TTTypes, TEKTypes,
     TTCommon, CommDlg, TTDialog;
{$ELSE}
uses WinTypes, WinProcs, WObjects, Win31, Strings, TTTypes, TEKTypes,
     TTCommon, CommDlg, TTDialog;
{$ENDIF}

procedure TEKInit(tk: PTEKVar; ts: PTTSet); export;
var
  i: integer;
begin
with tk^ do begin

  if (ts^.HideTitle=0) and (ts^.PopupMenu=0) then
  begin
    MainMenu := LoadMenu(HInstance, PChar(id_TEKMenu));
    EditMenu := GetSubMenu(MainMenu,1);
  end
  else
    MainMenu := 0;

  MemDC := 0;
  HBits := 0;
  Pen := 0;
  MemPen := 0;
  ps := ps_solid;
  BackGround := 0;
  MemBackGround := 0;
  for i := 0 to 3 do
    TEKFont[i] := 0;
  ScreenHeight := 450;
  ScreenWidth := 600;
  AdjustSize := false;
  ScaleFont := false;
  TextSize := 0;

  FillChar(TEKLogFont, SizeOf(TLogFont), #0);
  with TEKLogFont do
  begin
    lfHeight        := ts^.TEKFontSize.y;  {Font Height}
    lfWidth         := ts^.TEKFontSize.x;  {Font Width}
    lfCharSet       := ts^.TEKFontCharSet;  {Character Set}
    StrCopy(lfFaceName, @ts^.TEKFont);
  end;

  MoveFlag := true;

  ParseMode := ModeFirst;
  DispMode := IdAlphaMode;
  Drawing := false;

  RubberBand := false;
  Select := false;
  ButtonDown := false;

  GIN := false;
  CrossHair := false;
  IgnoreCount := 0;
  GINX := 0;
  GINY := 0;

end;
end;

procedure Log1Byte(cv: PComVar; b: byte);
begin
  with cv^ do begin
    PChar(LogBuf)[LogPtr] := char(b);
    inc(LogPtr);
    if LogPtr>=InBuffSize then LogPtr := LogPtr-InBuffSize;
    if LCount>=InBuffSize then
    begin
      LCount := InBuffSize;
      LStart := LogPtr;
    end
    else inc(LCount);
  end;
end;

procedure ToggleCrossHair(tk: PTEKVar; ts: PTTSet; OnFlag: bool);
var
  DC: HDC;
  TempPen, OldPen: HPen;
begin
with tk^ do begin
  if CrossHair=OnFlag then exit;
  DC := GetDC(HWin);
  TempPen := CreatePen(ps_solid, 1, ts^.TEKColor[0]);
  OldPen := SelectObject(DC,TempPen);
  SetROP2(DC, r2_Not);
  MoveTo(DC,GINX,0);
  LineTo(DC,GINX,ScreenHeight-1);
  MoveTo(DC,0,GINY);
  LineTo(DC,ScreenWidth-1,GINY);
  SelectObject(DC,OldPen);
  DeleteObject(TempPen);
  ReleaseDC(HWin,DC);
  CrossHair := OnFlag;
end;
end;


procedure TEKCaretOn(tk: PTEKVar; ts: PTTSet);
begin
with tk^ do begin
  if not Active then exit;
  if DispMode = IdAlphaMode then
  begin
    if ts^.CursorShape=IdHCur then
      SetCaretPos(CaretX,CaretY+FontHeight-CurWidth)
    else
      SetCaretPos(CaretX,CaretY);

    while CaretStatus > 0 do
    begin
      ShowCaret(HWin);
      Dec(CaretStatus);
    end;
  end;
end;
end;

procedure TEKCaretOff(tk: PTEKVar);
begin
with tk^ do begin
  if not Active then exit;
  if CaretStatus = 0 then
  begin
    HideCaret(HWin);
    inc(CaretStatus);
  end;
end;
end;

procedure ChangeCaret(tk: PTEKVar; ts: PTTSet); export;
begin
with tk^ do begin
  case ts^.CursorShape of
    IdHCur: CreateCaret(HWin, 0, FontWidth, CurWidth);
    IdVCur: CreateCaret(HWin, 0, CurWidth, FontHeight);
  else
    CreateCaret(HWin, 0, FontWidth, FontHeight);
  end;
  CaretStatus := 1;
  TEKCaretOn(tk,ts);
end;
end;


procedure SwitchRubberBand(tk: PTEKVar; ts: PTTSet; OnFlag: bool);
var
  DC: HDC;
  TempPen, OldPen: HPen;
  OldMemRop: integer;
  OldMemBrush: HBrush;
begin
with tk^ do begin
  if RubberBand=OnFlag then exit;

  TempPen := CreatePen(ps_Dot, 1, ts^.TEKColor[0]);
  DC := GetDC(HWin);
  SetBkMode(DC,1);
  SelectObject(DC,GetStockObject(hollow_Brush));
  OldPen := SelectObject(DC,TempPen);
  SetROP2(DC, r2_Not);
  Rectangle(DC,SelectStart.X,SelectStart.Y,SelectEnd.X,SelectEnd.Y);
  SelectObject(DC,OldPen);
  ReleaseDC(HWin,DC);
  DeleteObject(TempPen);

  TempPen := CreatePen(ps_Dot, 1, MemForeColor);
  OldMemBrush := SelectObject(MemDC,GetStockObject(hollow_Brush));
  SelectObject(MemDC,TempPen);
  OldMemRop := SetROP2(MemDC, r2_XorPen);
  Rectangle(MemDC,SelectStart.X,SelectStart.Y,SelectEnd.X,SelectEnd.Y);
  SelectObject(MemDC,OldMemBrush);
  SetRop2(MemDC,OldMemRop);
  SelectObject(MemDC,MemPen);
  DeleteObject(TempPen);     

  RubberBand := OnFlag;
end;
end;

procedure ChangeTextSize(tk: PTEKVar; ts: PTTSet);
var
  i: integer;
begin
with tk^ do begin
  SelectObject(MemDC,TEKFont[TextSize]);
  FontWidth := FW[TextSize];
  FontHeight := FH[TextSize];

  for i := 0 to 1 do
    Dx[i] := FontWidth;

  if Active then ChangeCaret(tk, ts);
end;
end;

procedure TEKResizeWindow(tk: PTEKVar; ts: PTTSet; W, H: Integer); export;
var
  i,Height,Width: integer;
  Metrics: TTextMetric;
  TempDC: HDC;
begin
with tk^ do begin
  if tk^.Select then SwitchRubberband(tk,ts,false);
  Select := false;

  {Delete old MemDC}
  if MemDC<>0 then
  begin
    SelectObject(MemDC, OldMemFont);
    SelectObject(MemDC, OldMemBmp);
    SelectObject(MemDC, OldMemPen);
    DeleteDC(MemDC);
  end;

  {Delete old fonts}
  for i :=0 to 3 do
    if TEKFont[i]<>0 then DeleteObject(TEKFont[i]);

  {Delete old bitmap}
  if HBits <> 0 then DeleteObject(HBits);

  {Delete old pen}
  if Pen <> 0 then DeleteObject(Pen);
  if MemPen <> 0 then DeleteObject(MemPen);

  {Delete old brush}
  if BackGround <> 0 then DeleteObject(BackGround);
  if MemBackGround <> 0 then DeleteObject(MemBackGround);

  {get DC}
  TempDc := GetDC(HWin);
  MemDC := CreateCompatibleDC(TempDc);

  {Create standard size font}
  if ScaleFont then
  begin
    TEKLogFont.lfHeight := round(ScreenHeight / 35);
    TEKLogFont.lfWidth := round(ScreenWidth / 74);
  end;

  with TEKLogFont do
  begin
    lfWeight := FW_Normal;
    lfItalic := 0;
    lfUnderLine := 0;
    lfStrikeOut := 0;
    lfOutPrecision  := Out_Character_Precis;
    lfClipPrecision := Clip_Character_Precis;
    lfQuality       := Default_Quality;
    lfPitchAndFamily:= Fixed_Pitch or FF_Modern;
  end;

  TEKFont[0] := CreateFontIndirect(TEKLogFont);
  {Check standard font size}
  SelectObject(TempDC,TEKFont[0]);
  GetTextMetrics(TempDC, Metrics);
  FW[0] := Metrics.tmAveCharWidth;
  FH[0] := Metrics.tmHeight;
 
  if not ScaleFont then
  begin
    ScreenHeight := FH[0]*35;
    ScreenWidth := FW[0]*74;
    Width := round(ScreenHeight/3120*4096);
    if ScreenWidth < Width then
      ScreenWidth := Width;
  end;

  Height := TEKLogFont.lfHeight;
  Width := TEKLogFont.lfWidth;

  {second font}
  TEKLogFont.lfHeight := round(ScreenHeight/38);
  TEKLogFont.lfWidth := round(ScreenWidth/80);
  TEKFont[1] := CreateFontIndirect(TEKLogFont);
  SelectObject(TempDC,TEKFont[1]);
  GetTextMetrics(TempDC, Metrics);
  FW[1] := Metrics.tmAveCharWidth;
  FH[1] := Metrics.tmHeight;

  {third font}
  TEKLogFont.lfHeight := round(ScreenHeight/58);
  TEKLogFont.lfWidth := round(ScreenWidth/121);
  TEKFont[2] := CreateFontIndirect(TEKLogFont);
  SelectObject(TempDC,TEKFont[2]);
  GetTextMetrics(TempDC, Metrics);
  FW[2] := Metrics.tmAveCharWidth;
  FH[2] := Metrics.tmHeight;

  {forth font}
  TEKLogFont.lfHeight := round(ScreenHeight/64);
  TEKLogFont.lfWidth := round(ScreenWidth/133);
  TEKFont[3] := CreateFontIndirect(TEKLogFont);
  SelectObject(TempDC,TEKFont[3]);
  GetTextMetrics(TempDC, Metrics);
  FW[3] := Metrics.tmAveCharWidth;
  FH[3] := Metrics.tmHeight;

  OldMemFont := SelectObject(MemDC,TEKFont[TextSize]);
  FontWidth := FW[TextSize];
  FontHeight := FH[TextSize];
  for i := 0 to 1 do
    Dx[i] := FontWidth;

  TEKLogFont.lfHeight := Height;
  TEKLogFont.lfWidth := Width;

  if ts^.TEKColorEmu>0 then
    HBits := CreateCompatibleBitmap(TempDC,ScreenWidth,ScreenHeight)
  else
    HBits := CreateBitmap(ScreenWidth, ScreenHeight, 1, 1, nil);

  OldMemBmp := SelectObject(MemDC, HBits);

  TextColor := ts^.TEKColor[0];
  if ts^.TEKColorEmu>0 then
  begin
    MemForeColor := ts^.TEKColor[0];
    MemBackColor := ts^.TEKColor[1];
  end
  else begin
    MemForeColor := RGB(0,0,0);
    MemBackColor := RGB(255,255,255);
  end;
  MemTextColor := MemForeColor;

  SetTextColor(MemDC, MemTextColor);
  SetBkColor(MemDC,MemBackColor);
  SetBkMode(MemDC, 1);

  BackGround := CreateSolidBrush(ts^.TEKColor[1]);
  MemBackGround := CreateSolidBrush(MemBackColor);

  PenColor := ts^.TEKColor[0];
  Pen := CreatePen(ps,1,PenColor);

  MemPenColor := MemForeColor;
  MemPen := CreatePen(ps,1,MemPenColor);
  OldMemPen := SelectObject(MemDC, MemPen);

  ReleaseDC(HWin,TempDC);

  R.left := 0;
  R.right := ScreenWidth;
  R.top := 0;
  R.bottom := ScreenHeight;
  FillRect(MemDC,R,MemBackGround);
  InvalidateRect(HWin, @R, true);

  DispMode := IdAlphaMode;
  HiY := 0;
  Extra := 0;
  LoY := 0;
  HiX := 0;
  LoX := 0;
  CaretX := 0;
  CaretOffset := 0;
  CaretY := 0;
  if Active then ChangeCaret(tk,ts);

  GetClientRect(HWin,R);
  Width := W + ScreenWidth - R.Right + R.Left;
  Height := H + ScreenHeight - R.Bottom + R.Top;
  if (Width<>W) or (Height<>H) then
  begin
    AdjustSize := true;
    SetWindowPos(HWin,HWND_TOP,0,0,Width,Height,SWP_NOMOVE);
  end;

  ScaleFont := false;
end;
end;

procedure BackSpace(tk: PTEKVar; cv: PComVar);
begin
with tk^ do begin
  CaretX := CaretX - FontWidth;
  if CaretX < 0 then CaretX := 0;
  if cv^.HLogBuf<>0 then Log1Byte(cv,BS);
end;
end;

procedure LineFeed(tk: PTEKVar; cv: PComVar);
var
  CursorY: integer;
begin
with tk^ do begin
  CaretY := CaretY + FontHeight;
  if CaretY > ScreenHeight - FontHeight then
  begin
    CaretY := 0;
    if (CaretOffset=0) or (CaretX < CaretOffset) then
    begin
      CaretOffset := ScreenWidth div 2;
      CaretX := CaretX + CaretOffset;
      if CaretX >= ScreenWidth then CaretX := CaretOffset;
    end
    else begin
      CaretX := CaretX - CaretOffset;
      CaretOffset := 0;
    end;
  end;
  if cv^.HLogBuf<>0 then Log1Byte(cv,LF);
end;
end;

procedure CarriageReturn(tk: PTEKVar; cv: PComVar);
begin
with tk^ do begin
  CaretX := CaretOffset;
  if (cv^.HLogBuf<>0) then Log1Byte(cv,CR);
end;
end;

procedure Tab(tk: PTEKVar; cv: PComVar);
begin
with tk^ do begin
  if cv^.HLogBuf<>0 then Log1Byte(cv,HT);
  CaretX := CaretX + FontWidth*8;
  if (CaretX>=ScreenWidth) then
  begin
    CarriageReturn(tk,cv);
    LineFeed(tk,cv);
  end;
end;
end;

procedure EnterVectorMode(tk: PTEKVar);
begin
with tk^ do begin
  MoveFlag := true;
  if DispMode = IdAlphaMode then
    CaretY := CaretY + FontHeight - 1;
  Drawing := false;
  DispMode := IdVectorMode;
end;
end;

procedure EnterPlotMode(tk: PTEKVar);
begin
with tk^ do begin
  if DispMode = IdAlphaMode then
  begin
    CaretY := CaretY + FontHeight - 1;
    CaretOffset := 0;
  end;
  PlotX := Trunc(CaretX / ScreenWidth * ViewSizeX);
  PlotY := Trunc((ScreenHeight-1-CaretY) / ScreenHeight * ViewSizeY);
  DispMode := IdPlotMode;
  JustAfterRS := true;
end;
end;

procedure EnterAlphaMode(tk: PTEKVar);
begin
with tk^ do begin
  if (DispMode = IdVectorMode) or
     (DispMode = IdPlotMode) then
  begin
    CaretY := CaretY - FontHeight + 1;
    CaretOffset := 0;
  end;
  DispMode := IdAlphaMode;
  Drawing := false;
end;
end;

procedure EnterUnknownMode(tk: PTEKVar);
begin
with tk^ do begin
  if (DispMode = IdVectorMode) or
     (DispMode = IdPlotMode) then
  begin
    CaretY := CaretY - FontHeight + 1;
    CaretOffset := 0;
  end;
  DispMode := IdUnknownMode;
end;
end;

procedure TEKMoveTo(tk: PTEKVar; ts: PTTSet; X, Y: integer; Draw: boolean);
var
  DC: HDC;
  OldPen: HPen;
begin
with tk^ do begin
  if Draw then
  begin
    DC := GetDC(HWin);
    OldPen := SelectObject(DC,Pen);
    SetBkColor(DC,ts^.TEKColor[1]);
    SetBkMode(DC, 1);
    if (CaretX=X) and (CaretY=Y)
    then begin
      SetPixel(DC,X,Y,PenColor);
      SetPixel(MemDC,X,Y,MemPenColor);
    end
    else begin
      MoveTo(DC,CaretX,CaretY);
      LineTo(DC,X,Y);
      MoveTo(MemDC,CaretX,CaretY);
      LineTo(MemDC,X,Y);
    end;
    SelectObject(DC,OldPen);
    ReleaseDC(HWin,DC);
  end;
  CaretX := X;
  CaretY := Y;
end;
end;

procedure Draw(tk: PTEKVar; cv: PComVar; ts: PTTSet; b: byte);
var
  X, Y: integer;
begin
with tk^ do begin
  if not Drawing then
  begin
    LoXReceive := false;
    LoCount := 0;
    Drawing := true;
  end;

  case b of
    $00..$1f: begin
                CommInsert1Byte(cv,b);
                exit;
              end;
    $20..$3f: if LoCount=0 then HiY := b - $20
                           else HiX := b - $20;
    $40..$5f: begin
                LoX := b - $40;
                LoXReceive := true;
              end;
    $60..$7f: begin
                LoA := LoB;
                LoB := b - $60;
                inc(LoCount);
              end;
  end;

  if not LoXReceive then exit;

  Drawing := false;
  ParseMode := ModeFirst;
  if LoCount>0 then LoY := LoB;
  if LoCount>1 then Extra := LoA;

  X := Trunc((HiX*128+LoX*4+(Extra and 3)) / ViewSizeX * ScreenWidth);
  Y := ScreenHeight-1-Trunc((HiY*128+LoY*4+Extra div 4) / ViewSizeY * ScreenHeight);
  TEKMoveTo(tk,ts,X,Y,not MoveFlag);
  MoveFlag := false;
end;
end;

procedure Plot(tk: PTEKVar; ts: PTTSet; b: byte);
var
  X, Y: integer;
begin
with tk^ do begin
  X := 0;
  Y := 0;
  case b of
    ord('A'): X := 4;
    ord('B'): X := -4;
    ord('D'): Y := 4;
    ord('E'): begin
                X := 4;
                Y := 4;
              end;
    ord('F'): begin
                X := -4;
                Y := 4;
              end;
    ord('H'): Y := -4;
    ord('I'): begin
                X := 4;
                Y := -4;
              end;  
    ord('J'): begin
                X := -4;
                Y := -4;
              end;
  else
    exit;
  end;
  PlotX := PlotX + X;
  if PlotX < 0 then PlotX := 0;
  if PlotX >= ViewSizeX then PlotX := ViewSizeX - 1;
  PlotY := PlotY + Y;
  if PlotY < 0 then PlotY := 0;
  if PlotY >= ViewSizeY then PlotY := ViewSizeY - 1;
  X := Trunc(PlotX / ViewSizeX * ScreenWidth);
  Y := ScreenHeight-1-Trunc(PlotY / ViewSizeY * ScreenHeight);
  TekMoveTo(tk,ts,X,Y,PenDown);
end;
end;


procedure DispChar(tk: PTEKVar; cv: PComVar; b: byte);
begin
with tk^ do begin
  R.left := CaretX;
  R.Right := CaretX + FontWidth;
  R.top := CaretY;
  R.bottom := CaretY+FontHeight;
  ExtTextOut(MemDc,CaretX,CaretY,0,@R,@b,1,@Dx);
  InvalidateRect(HWin,@R,false);
  CaretX := R.Right;

  if cv^.HLogBuf<>0 then Log1Byte(cv,b);

  if (CaretX>ScreenWidth-FontWidth) then
  begin
    CarriageReturn(tk,cv);
    LineFeed(tk,cv);
  end;
end;
end;


procedure ParseFirst(tk: PTEKVar; ts: PTTSet; cv: PComVar; b: byte);
begin
with tk^ do begin
  if DispMode = IdAlphaMode then
    case b of
      $00..$06:;
      $07: MessageBeep(0);
      $08: BackSpace(tk,cv);
      $09: Tab(tk,cv);
      $0a: LineFeed(tk,cv);
      $0b..$0c:;
      $0d: begin
             CarriageReturn(tk,cv);
             if ts^.CRReceive=IdCRLF then
               CommInsert1Byte(cv,$0A);
           end;
      $0e..$17:;
      $18: if ts^.AutoWinSwitch>0 then ChangeEmu := IdVT; {Enter VT Mode}
      $19..$1a:;
      $1b: begin
             ParseMode := ModeEscape;
             exit;
           end;
      $1c: EnterUnknownMode(tk);
      $1d: EnterVectorMode(tk);
      $1e: EnterPlotMode(tk);
      $1f: ;
    else
      if Drawing then Draw(tk,cv,ts,b)
                 else DispChar(tk,cv,b);
    end
  else if DispMode=IdVectorMode then
    case b of
      $00..$06:;
      $07: if MoveFlag then MoveFlag := false
                       else MessageBeep(0);
      $08..$0c: ;
      $0d: begin
             CommInsert1Byte(cv,$0d);
             EnterAlphaMode(tk);  {EnterAlphaMode}
           end;
      $0e..$17:;
      $18: if ts^.AutoWinSwitch>0 then ChangeEmu := IdVT; {Enter VT Mode}
      $19..$1a:;
      $1b: begin
             ParseMode := ModeEscape;
             exit;
           end;
      $1c: EnterUnknownMode(tk);
      $1d: EnterVectorMode(tk);
      $1e: EnterPlotMode(tk);
      $1f: EnterAlphaMode(tk);  {EnterAlphaMode}
    else
        Draw(tk,cv,ts,b);
    end
  else if DispMode=IdPlotMode then
    case b of
      $00..$06:;
      $07: MessageBeep(0);
      $08..$0c: ;
      $0d: begin
             CommInsert1Byte(cv,$0d);
             EnterAlphaMode(tk);  {EnterAlphaMode}
           end;
      $0e..$17:;
      $18: if ts^.AutoWinSwitch>0 then ChangeEmu := IdVT; {Enter VT Mode}
      $19..$1a:;
      $1b: begin
             ParseMode := ModeEscape;
             exit;
           end;
      $1c: EnterUnknownMode(tk);
      $1d: EnterVectorMode(tk);
      $1e: EnterPlotMode(tk);
      $1f: EnterAlphaMode(tk);  {EnterAlphaMode}
    else
      if JustAfterRS then
      begin
        if b=$20 then
          PenDown := false
        else if b=ord('P') then
          PenDown := true
        else
          EnterUnknownMode(tk);
        JustAfterRS := false;
      end
      else             
        Plot(tk,ts,b);
    end
  else
    case b of
      $1f: EnterAlphaMode(tk);
    end;  

  ParseMode := ModeFirst;
end;
end;

{$i tekesc.pas}

function TEKParse(tk: PTEKVar; ts: PTTSet; cv: PComVar): integer; export;
var
  f: boolean;
  c: integer;
  b: byte;

  DummyMsg: TMsg;
begin
with tk^ do begin
  if ButtonDown then
  begin
    TEKParse := 0;
    exit
  end;

  ChangeEmu := 0;
  f := true;

  repeat
    c := CommRead1Byte(cv,@b);

    if c > 0 then
    begin
      if IgnoreCount <= 0 then
      begin
        if f then
        begin
        TEKCaretOff(tk);
          f := false;
          if GIN then ToggleCrossHair(tk,ts,false);
          if RubberBand then SwitchRubberBand(tk,ts,false);
        end;

        case ParseMode of
          ModeFirst: ParseFirst(tk,ts,cv,b);
          ModeEscape: Escape(tk,ts,cv,b);
          ModeSelectCode: SelectCode(tk,ts,b);
          Mode2OC: TwoOpCode(tk,ts,cv,b);
        else
          begin
            ParseMode := ModeFirst;
            ParseFirst(tk,ts,cv,b);
          end;
        end;
        PeekMessage(DummyMsg, 0, 0, 0, pm_NoRemove);
      end
      else Dec(IgnoreCount);
    end;
  until (c=0) or (ChangeEmu>0);

  ToggleCrossHair(tk,ts,GIN);
  if not f then
  begin
    TEKCaretOn(tk,ts);
    SwitchRubberBand(tk,ts,Select);
  end;

  if ChangeEmu > 0 then ParseMode := ModeFirst;
  TEKParse := ChangeEmu;
end;
end;


procedure ReportGIN(tk: PTEKVar; ts: PTTSet; cv: PComVar; KeyCode: byte); export;
var
  Code: array[0..10] of byte;
  X, Y: integer;
begin
with tk^ do begin
  ToggleCrossHair(tk,ts,false);
  X := Trunc(GINX/ScreenWidth*ViewSizeX);
  Y := Trunc((1-(GINY+1)/ScreenHeight)*ViewSizeY);
  Code[0] := KeyCode;
  Code[1] := (X shr 7) + 32;
  Code[2] := ((X shr 2) and $1f) + 32;
  Code[3] := (Y shr 7) + 32;
  Code[4] := ((Y shr 2) and $1f) + 32;
  Code[5] := $0d;
  GIN := false;
  ReleaseCapture;
  CommBinaryOut(cv, @Code[0],6);
  IgnoreCount := 6;
end;
end;

procedure TEKPaint(tk: PTEKVar; ts: PTTSet; PaintDC: HDC; PaintInfo: PPaintStruct); export;
var
  X,Y,W,H: integer;
begin
with tk^ do begin
  if PaintInfo^.fErase then
    FillRect(PaintDC, PaintInfo^.rcPaint,Background);

  if GIN then ToggleCrossHair(tk,ts,false);
  if Select then SwitchRubberBand(tk,ts,false);
  X := PaintInfo^.rcPaint.left;
  Y := PaintInfo^.rcPaint.top;
  W := PaintInfo^.rcPaint.right - X;
  H := PaintInfo^.rcPaint.bottom - Y;
  SetTextColor(PaintDC, ts^.TEKColor[0]);
  SetBkColor(PaintDC,  ts^.TEKColor[1]);
  BitBlt(PaintDC,X,Y,W,H,MemDC,X,Y,SrcCopy);
  SwitchRubberBand(tk,ts,Select);
  if GIN then ToggleCrossHair(tk,ts,true);
end;
end;

procedure TEKInitMenu(tk: PTEKVar; cv: PComVar; SubMenu: HMenu); export;
begin
with tk^ do begin
  if SubMenu=EditMenu then
  begin
    if Select then
      EnableMenuItem(EditMenu,cm_EditCopy,mf_ByCommand or mf_Enabled)
    else
      EnableMenuItem(EditMenu,cm_EditCopy,mf_ByCommand or mf_Grayed);

    if cv^.Ready and
       (IsClipboardFormatAvailable(cf_Text) or
        IsClipboardFormatAvailable(cf_OEMText)) then
    begin
      EnableMenuItem(EditMenu,cm_EditPaste,mf_ByCommand or mf_Enabled);
      EnableMenuItem(EditMenu,cm_EditPasteCR,mf_ByCommand or mf_Enabled);
    end
    else begin
      EnableMenuItem(EditMenu,cm_EditPaste,mf_ByCommand or mf_Grayed);
      EnableMenuItem(EditMenu,cm_EditPasteCR,mf_ByCommand or mf_Grayed);
    end;
  end;
end;
end;

procedure TEKWMLButtonDown(tk: PTEKVar; ts: PTTSet; cv: PComVar; Popup: BOOL; pos: TPoint); export;
var
  b: byte;
begin
with tk^ do begin
  if GIN then
  begin
    b := ts^.GINMouseCode and $ff;
    ReportGIN(tk,ts,cv,b);
    exit;
  end;

  if ButtonDown then exit;

  if Popup then
  begin
    PopupMenu := LoadMenu(HInstance, PChar(id_TEKMenu));
    EditMenu := GetSubMenu(PopupMenu,1);
    TEKInitMenu(tk,cv,EditMenu);
    PopupBase := CreatePopupMenu;
    AppendMenu(PopupBase, MF_STRING or MF_ENABLED or MF_POPUP,
               GetSubMenu(PopupMenu,0), '&File');
    AppendMenu(PopupBase, MF_STRING or MF_ENABLED or MF_POPUP,
               EditMenu, '&Edit');
    AppendMenu(PopupBase, MF_STRING or MF_ENABLED or MF_POPUP,
               GetSubMenu(PopupMenu,2), '&Setup');
    AppendMenu(PopupBase, MF_STRING or MF_ENABLED,
               cm_VTWin, 'VT&Win');
    AppendMenu(PopupBase, MF_STRING or MF_ENABLED or MF_POPUP,
               GetSubMenu(PopupMenu,4), '&Help');
    exit;
  end;

  {Capture mouse}
  SetCapture(HWin);

  {Is the position in client area?}
  if (pos.x>=0) and (pos.x<ScreenWidth) and
     (pos.y>=0) and (pos.y<ScreenHeight) then
  begin
    SwitchRubberBand(tk,ts,false);
    Select := false;

    SelectStart.x := pos.x;
    SelectStart.y := pos.y;
    SelectEnd := SelectStart;
    ButtonDown := true;
  end;

end;
end;

procedure TEKWMLButtonUp(tk: PTEKVar; ts: PTTSet); export;
var
  X: integer;
begin
with tk^ do begin
  ReleaseCapture;
  ButtonDown := false;
  if (Abs(SelectEnd.Y-SelectStart.Y)>2) and
     (Abs(SelectEnd.X-SelectStart.X)>2) then
  begin
    if SelectStart.X>SelectEnd.X then
    begin
      X := SelectEnd.X;
      SelectEnd.X := SelectStart.X; 
      SelectStart.X := X;
    end;
    if SelectStart.Y>SelectEnd.Y then
    begin
      X := SelectEnd.Y;
      SelectEnd.Y := SelectStart.Y; 
      SelectStart.Y := X;
    end;
    Select := true;
  end
  else begin
    SwitchRubberBand(tk,ts,false);
    Select := false;
  end;
end;
end;

procedure TEKWMMouseMove(tk: PTEKVar; ts: PTTSet; p: TPoint); export;
var
  X, Y: integer;
begin
with tk^ do begin
  if (not ButtonDown) and (not GIN) then exit;
  {get position}
  X := p.x + 1;
  Y := p.y + 1;

  { if out of client area, force into client area}
  if X<0 then X := 0
  else if X>ScreenWidth then X := ScreenWidth - 1;
  if Y<0 then Y := 0
  else if Y>ScreenHeight then Y := ScreenHeight - 1;

  if GIN then
  begin
    ToggleCrossHair(tk,ts,false);
    GINX := X;
    GINY := Y;
    ToggleCrossHair(tk,ts,true);
  end
  else begin
    SwitchRubberBand(tk,ts,false);
    SelectEnd.X := X + 1;
    SelectEnd.Y := Y + 1;
    SwitchRubberBand(tk,ts,true);
  end;  

  if GIN then SetCapture(HWin);
end;
end;

procedure TEKWMSize(tk: PTEKVar; ts: PTTSet; W, H, cx, cy: Integer); export;
var
  Width, Height: integer;
begin
with tk^ do begin

  Width := cx;
  Height := cy;

  if (ScreenWidth=Width) and
     (ScreenHeight=Height) then
  begin
    AdjustSize := FALSE;
    exit;
  end;

  if AdjustSize then
  begin
    Width := W + ScreenWidth - Width;
    Height := H + ScreenHeight - Height;
    SetWindowPos(HWin,HWND_TOP,0,0,Width,Height,SWP_NOMOVE);
  end
  else begin
    ScreenWidth := Width;
    ScreenHeight := Height;
    ScaleFont := true;
    TEKResizeWindow(tk,ts,W,H);
  end;
end;
end;

procedure CopyToClipboard(tk: PTEKVar; ts: PTTSet; x, y, Width,Height: integer);
var
  CopyDC: HDC;
  CopyBitmap: HBitmap;
begin
with tk^ do begin
  if Select then SwitchRubberBand(tk,ts,false);
  TEKCaretOff(tk);
  if OpenClipBoard(HWin) and EmptyClipBoard then
  begin
    { Create the new bitmap }
    CopyDC := CreateCompatibleDC(MemDC);
    CopyBitmap := CreateCompatibleBitmap(MemDC, Width, Height);
    CopyBitmap := SelectObject(CopyDC, CopyBitmap);
    BitBlt(CopyDC, 0, 0, Width, Height, MemDC, x, y, SrcCopy);
    CopyBitmap := SelectObject(CopyDC, CopyBitmap);
    { Transfer the new bitmap to the clipboard }
    SetClipBoardData(cf_Bitmap, CopyBitmap);
  end;

  CloseClipBoard;
  DeleteDC(CopyDC);

  TEKCaretOn(tk,ts);
  SwitchRubberBand(tk,ts,Select);
end;
end;

procedure TEKCMCopy(tk: PTEKVar; ts: PTTSet); export;
var
  x, y: integer;
begin
with tk^ do begin
  if not Select then exit;

  if SelectStart.X < SelectEnd.X then x := SelectStart.X
                                 else x := SelectEnd.X;
  if SelectStart.Y < SelectEnd.Y then y := SelectStart.Y
                                 else y := SelectEnd.Y;
  {copy selected area to clipboard}
  CopyToClipboard(tk, ts, x,y,abs(SelectEnd.X-SelectStart.X),
                  abs(SelectEnd.Y-SelectStart.Y));                          
end;
end;

procedure TEKCMCopyScreen(tk: PTEKVar; ts: PTTSet); export;
begin
with tk^ do begin
  {copy fullscreen to clipboard}
  CopyToClipboard(tk, ts, 0, 0, ScreenWidth, ScreenHeight);
end;
end;

procedure TEKPrint(tk: PTEKVar; ts: PTTSet; PrintDC: HDC; SelFlag: longint); export;
var
  PPI: TPoint;
  Mergin: TRect;
  Caps: integer;
  PrnWidth, PrnHeight: integer;
  MemWidth, MemHeight: integer;
  PrintStart, PrintEnd: TPoint;
begin
with tk^ do begin

  if PrintDC = 0 then exit;

  if (SelFlag and PD_Selection) <> 0 then
  begin
    {print selection}
    PrintStart := SelectStart;
    PrintEnd := SelectEnd;
  end
  else begin
    {print current page} 
    PrintStart.X := 0;
    PrintStart.Y := 0;
    PrintEnd.X := ScreenWidth;
    PrintEnd.Y := ScreenHeight;
  end;

  Caps := GetDeviceCaps(PrintDC,RasterCaps);
  if (Caps and rc_BitBlt)<>rc_BitBlt then
  begin
    messagebox(HWin,'Printer dose not support graphics','Tera Term: Error',mb_IconExclamation);
    exit;
  end;
  if Active then TEKCaretOff(tk);
  if RubberBand then SwitchRubberBand(tk,ts,false);

  MemWidth := PrintEnd.X - PrintStart.X;
  MemHeight := PrintEnd.Y - PrintStart.Y;
  if (MemWidth=0) or (MemHeight=0) then exit;

  StartPage(PrintDC);


  PPI.X := GetDeviceCaps(PrintDC,LogPixelSX);
  PPI.Y := GetDeviceCaps(PrintDC,LogPixelSY);

  Mergin.left := round(PPI.X*0.25); {horizontal mergin: about 1/4 inch}
  Mergin.right := GetDeviceCaps(PrintDC,HorzRes) - Mergin.left;
  Mergin.top := round(PPI.Y*1.0); {virtical mergin: about 1 inch}
  Mergin.bottom := GetDeviceCaps(PrintDC,VertRes) - Mergin.top;

  if (Caps and rc_StretchBlt) = rc_StretchBlt then
  begin
    PrnWidth := round(MemWidth/GetDeviceCaps(MemDC,LogPixelSX)*PPI.X);
    PrnHeight := round(MemHeight/GetDeviceCaps(MemDC,LogPixelSY)*PPI.Y);
    StretchBlt(PrintDC, Mergin.left, Mergin.top, PrnWidth, PrnHeight,
               MemDC, PrintStart.X, PrintStart.Y, MemWidth, MemHeight, SrcCopy);
  end
  else
    BitBlt(PrintDC, Mergin.left, Mergin.top, MemWidth, MemHeight,
               MemDC, PrintStart.X, PrintStart.Y, SrcCopy);

  EndPage(PrintDC);

  SwitchRubberBand(tk,ts,Select);
end;
end;

procedure TEKClearScreen(tk: PTEKVar); export;
begin
with tk^ do begin
  GetClientRect(HWin,R);
  FillRect(MemDC,R,MemBackground);
  InvalidateRect(HWin,@R,false);
  DispMode := IdAlphaMode;
  CaretX := 0;
  CaretOffset := 0;
  CaretY := 0;
  if Active then SetCaretPos(CaretX,CaretY);
end;
end;

procedure TEKSetupFont(tk: PTEKVar; ts: PTTSet); export;
var
  W, H: integer;
  Ok: bool;
begin
with tk^ do begin

  if not LoadTTDLG then exit;
  Ok := ChooseFontDlg(HWin,@TEKLogFont,nil);
  FreeTTDLG;
  if not OK then exit;

  StrCopy(ts^.TEKFont,TEKLogFont.lfFaceName);
  ts^.TEKFontSize.x := TEKLogFont.lfWidth;
  ts^.TEKFontSize.y := TEKLogFont.lfHeight;
  ts^.TEKFontCharSet := TEKLogFont.lfCharSet;

  GetWindowRect(HWin, R);
  W := R.right - R.left;
  H := R.bottom - R.top;
  TextSize := 0;
  TEKResizeWindow(tk,ts,W,H);
end;
end;

procedure TEKResetWin(tk: PTEKVar; ts: PTTSet; EmuOld: word);
var
  TmpDC: HDC;
begin
with tk^ do begin

  {Change caret shape}
  ChangeCaret(tk,ts);

  {Change display color}
  TmpDC := GetDC(HWin);

  if ts^.TEKColorEmu<>EmuOld then
  begin
    SelectObject(MemDC,OldMemBmp);
    if HBits<>0 then DeleteObject(HBits);
    if ts^.TEKColorEmu>0 then
      HBits := CreateCompatibleBitmap(TmpDC,ScreenWidth,ScreenHeight)
    else
      HBits := CreateBitmap(ScreenWidth, ScreenHeight, 1, 1, nil);
    OldMemBmp := SelectObject(MemDC, HBits);
    GetClientRect(HWin,R);
    FillRect(MemDC,R,MemBackground);
  end;

  ReleaseDC(HWin,TmpDC);

  TextColor := ts^.TEKColor[0];

  ps := ps_solid;
  PenColor := ts^.TEKColor[0];
  if Pen<>0 then DeleteObject(Pen);
  Pen := CreatePen(ps,1,PenColor);

  if BackGround<>0 then DeleteObject(BackGround);
  BackGround := CreateSolidBrush(ts^.TEKColor[1]);

  if ts^.TEKColorEmu>0 then
  begin
    MemForeColor := ts^.TEKColor[0];
    MemBackColor := ts^.TEKColor[1];
  end
  else begin
    MemForeColor := RGB(0,0,0);
    MemBackColor := RGB(255,255,255);
  end;
  MemTextColor := MemForeColor;
  MemPenColor := MemForeColor;

  SelectObject(MemDC, OldMemPen);
  if MemPen<>0 then DeleteObject(MemPen);
  MemPen := CreatePen(ps,1,MemPenColor);
  OldMemPen := SelectObject(MemDC,MemPen);

  if MemBackGround<>0 then DeleteObject(MemBackGround);
  MemBackGround := CreateSolidBrush(MemBackColor);

  SetTextColor(MemDC, MemTextColor);
  SetBkColor(MemDC, MemBackColor);

  if (ts^.TEKColorEmu>0) or
     (ts^.TEKColorEmu<>EmuOld) then
    TEKClearScreen(tk);

  InvalidateRect(HWin,nil,true);
end;
end;

procedure TEKSetupWinDlg(tk: PTEKVar; ts: PTTSet); export;
var
  OldEmu: word;
begin
with tk^ do begin

  ts^.OK := false;
  ts^.VTFlag := false;
  ts^.SampleFont := TEKFont[0];

  if not LoadTTDLG then exit;
  OldEmu := ts^.TEKColorEmu;
  SetupWin(HWin, ts);
  FreeTTDLG;

  if ts^.Ok then TEKResetWin(tk,ts,OldEmu);
end;
end;

procedure TEKRestoreSetup(tk: PTEKVar; ts: PTTSet); export;
var
  W, H: integer;
begin
with tk^ do begin
  {change window}
  StrCopy(TEKLogFont.lfFaceName,ts^.TEKFont);
  TEKLogFont.lfWidth := ts^.TEKFontSize.x;
  TEKLogFont.lfHeight := ts^.TEKFontSize.y;
  TEKLogFont.lfCharSet := ts^.TEKFontCharSet;
  GetWindowRect(HWin, R);
  W := R.right - R.left;
  H := R.bottom - R.top;
  TextSize := 0;
  TEKResizeWindow(tk,ts,W,H);
end;
end;

procedure TEKEnd(tk: PTEKVar); export;
var
  i: integer;
begin
with tk^ do begin
  if MemDC <> 0 then DeleteDC(MemDC);
  for i :=0 to 3 do
    if TEKFont[i]<>0 then DeleteObject(TEKFont[i]);
  if HBits <> 0 then DeleteObject(HBits);
  if Pen <> 0 then DeleteObject(Pen);
  if MemPen <> 0 then DeleteObject(MemPen);
  if BackGround <> 0 then DeleteObject(BackGround);
  if MemBackGround <> 0 then DeleteObject(MemBackGround);
end;
end;

procedure TEKChangeMenu(tk: PTEKVar; ts: PTTSet); export;
var
  SysMenu: HMenu;
begin
with tk^ do begin
  if (ts^.PopupMenu>0) and (MainMenu<>0)  or
     (ts^.PopupMenu=0) and (MainMenu=0) then
  begin
    if ts^.PopupMenu>0 then
    begin
      DestroyMenu(MainMenu);
      MainMenu := 0;
    end
    else begin
      MainMenu := LoadMenu(HInstance, PChar(id_TEKMenu));
      EditMenu := GetSubMenu(MainMenu,1);
    end;
    AdjustSize := true;
    SetMenu(HWin, MainMenu);
    DrawMenuBar(HWin);
  end;

  GetSystemMenu(HWin,true);
  if ts^.PopupMenu>0 then
  begin
    SysMenu := GetSystemMenu(HWin,false);
    AppendMenu(SysMenu, mf_Separator, 0, nil);
    AppendMenu(SysMenu, mf_String, cm_ShowMenubar, 'Show menu &bar');
  end;

end;
end;

exports

  TEKInit          index 1,
  TEKResizeWindow  index 2,
  ChangeCaret      index 3,
  TEKParse         index 4,
  ReportGIN        index 5,
  TEKPaint         index 6,
  TEKWMLButtonDown index 7,
  TEKWMLButtonUp   index 8,
  TEKWMMouseMove   index 9,
  TEKWMSize        index 10,
  TEKCMCopy        index 11,
  TEKCMCopyScreen  index 12,
  TEKInitMenu      index 13,
  TEKPrint         index 14,
  TEKClearScreen   index 15,
  TEKSetupFont     index 16,
  TEKSetupWinDlg   index 17,
  TEKRestoreSetup  index 18,
  TEKEnd           index 19,
  TEKChangeMenu    index 20;

begin
end.