{Tera Term
 Copyright(C) 1994-1996 T. Teranishi
 All rights reserved.}

{TERATERM.EXE, VT window}
unit VTWin;

interface
{$I teraterm.inc}

{$IFDEF Delphi}
uses Messages, WinTypes, WinProcs, OWindows, Strings, WinDos,
     CommDlg, ShellAPI, TEK, Telnet, TeraPrn, Keyboard,
     Buffer, Clipboard, Kanji, TTWinMan, FileSys, TTFTypes,
     TTDialog, IME, TTTypes, TTCommon, TTSetup, TTDDE,
     WinsockH, Ctl3d, TTLib, CommLib;
{$ELSE}
uses WinTypes, WinProcs, WObjects, Strings, Win31, WinDos,
     CommDlg, ShellAPI, TEK, Telnet, TeraPrn, Keyboard,
     Buffer, Clipboard, Kanji, TTWinMan, FileSys, TTFTypes,
     TTDialog, IME, TTTypes, TTCommon, TTSetup, TTDDE,
     WinsockH, Ctl3d, TTLib, CommLib;
{$ENDIF}

const
  Id_Menu = 100;
  IdVTIcon = 200;

  {mouse buttons}
  IdLeft = 0;
  IdMiddle = 1;
  IdRight = 2;

  CurWidth = 2;

  {Parsing modes}
  ModeFirst = 0;
  ModeESC   = 1;
  ModeDCS   = 2;
  ModeDCUserKey = 3;
  ModeSOS   = 4;
  ModeCSI   = 5;
  ModeXS    = 6;
  ModeDLE   = 7;
  ModeCAN   = 8;

  NParamMax = 16;
  IntCharMax = 5;

  DCUseCountLimit = 25;

type
  PVTScroller = ^VTSCroller;
  PVTWindow = ^VTWindow;

  VTScroller = object(TScroller)
    Parent: PVTWindow;
    constructor Init(TheWindow: PVTWindow; TheXUnit, TheYUnit: Integer;
                            TheXRange, TheYRange: Longint);
    procedure HScroll(ScrollRequest: Word; ThumbPos: Integer); virtual;
    procedure VScroll(ScrollRequest: Word; ThumbPos: Integer); virtual;
  end;

  VTWindow = object(TWindow)

    Param: array[1..NParamMax] of integer;
    NParam: integer;
    FirstPrm: boolean;
    IntChar: array[1..IntCharMax] of byte;
    ICount: integer;
    Prv: byte;

    ParseMode, SavedMode: integer;

    DC: HDC;
    CRTWidth, CRTHeight: integer;

    VTLogFont: TLogFont;
    OldFont: HFont;
    VTFont: array[0..AttrFontMask] of HFont;

    Dx, TabStops: array[0..255] of integer;
    NTabStops: integer;
    FontHeight, FontWidth, ScreenWidth, ScreenHeight: integer;
    AdjustSize, DontChangeSize: boolean;
    WinWidth, WinHeight, WinWidthOld, WinHeightOld: integer;
    WinOrgX, WinOrgY, NewOrgX, NewOrgY: integer;
    ScrollerChange: boolean;
    ScrollCount: integer;
    DCUseCount: integer;
    StrChangeStart, StrChangeCount: integer;

    CursorX, CursorY: integer;
    SavedCursor: TPoint;
    CursorTop, CursorBottom: integer;
    CharAttr, SavedAttr: byte;
    CharAttr2, SavedAttr2: byte;
    Special: boolean;
    Active, CompletelyVisible, Minimized: boolean;
    ChangeEmu: integer;

    {various modes for VT emulation}
    InsertMode: boolean;
    RelativeOrgMode, AutoWrapMode, Wrap: boolean;
    LFMode: boolean;

    R: TRect;

    ReverseColor: boolean;
    ANSIColor: array[0..7] of TColorRef;
    BackGround: HBrush;
    CaretStatus: integer;

    {mouse status}
    LButton, MButton, RButton, DblClk, TplClk: boolean;
    DblClkX, DblClkY: integer;

    {select parser flag ( ^[%!0 ) }
    SelectParserFlag: boolean;

    ESCFlag: boolean;
    JustAfterESC: boolean;

    Kanji: word;
    KanjiIn: boolean;
    EUCkanaIn, EUCsupIn: boolean;
    EUCcount: integer;
    {GL, GR code group}
    Glr: array[0..1] of integer;
    {G0, G1, G2, G3 code group}
    Gn: array[0..3] of integer;
    {single shift 2/3 flag}
    SSflag: boolean;
    {GL for single shift 2/3}
    GLtmp: integer;
    {JIS -> SJIS conversion flag}
    ConvJIS: boolean;

    FileMenu, TransMenu, EditMenu,
    SetupMenu, ControlMenu: HMenu;

    FirstPaint, DOption: Bool;

    constructor Init(Aparent: PWindowsObject);
    procedure DefWndProc(var Msg: TMessage); virtual;
    procedure SetupWindow; virtual;
    function GetClassName: PChar; virtual;
    procedure GetWindowClass(var AWndClass: TWndClass); virtual;

    procedure SetLogFont;
    procedure ChangeFont;
    procedure ChangeWindowSize(Nx, Ny: integer);
    procedure ChangeTerminalSize(Nx, Ny: integer);
    procedure Scroll1Line;
    procedure ChangeCaret;
    procedure CaretOn;
    procedure CaretOff;

    procedure DispChar(b: byte);
    procedure DispKanji(b: byte);
    procedure MoveRight;
    procedure EraseKanji(LR: integer);

    procedure BackSpace;
    procedure CarriageReturn;
    procedure PrintCurrentLine(b: byte);
    procedure LineFeed(b: byte);
    procedure Tab;
    procedure SetTabStop;
    procedure CursorUpWithScroll;

    function Parse: integer;
    procedure ParseFirst(b: byte);
    procedure ParseControl(b: byte);
    procedure EscapeSequence(b: byte);
    procedure SaveCursor;
    procedure RestoreCursor;
    procedure ParseEscape(b: byte);
    procedure ControlSequence(b: byte);
    procedure ParseCS(b: byte);
    procedure DeviceControl(b: byte);
    procedure DCUserKey(b: byte);
    procedure IgnoreString(b: byte);
    procedure DispAllChar(b: integer);
    procedure AnswerTerminalType;
    procedure XSequence(b: byte);
    procedure DLESeen(b: byte);
    procedure CANSeen(b: byte);

    procedure ResetTerminal;
    procedure ResetCharSet;
    procedure ResetIME;
    procedure ClearBuffer;
    procedure MoveCursor(Xnew,Ynew: integer);
    procedure UpdateStr;
    procedure UpdateRect(UpdateDC: HDC; XStart,YStart,XEnd,YEnd: integer);
    procedure SetupDC;
    procedure CountDCUse;
    procedure CountScroll;
    procedure UpdateScroll;
    procedure ScrollToCursor;
    function  CheckKanji(b: byte): boolean;

    procedure WMDestroy(var Msg: TMessage);
      virtual wm_First + wm_Destroy;
    procedure WMDropFiles(var Msg: TMessage);
      virtual wm_First + wm_DropFiles;
    procedure WMGetMinMaxInfo(var Msg: TMessage);
      virtual wm_First + wm_GetMinMaxInfo;
    procedure WMSysKeyDown(var Msg: TMessage);
      virtual wm_First + wm_SysKeyDown;
    procedure WMSysKeyUp(var Msg: TMessage);
      virtual wm_First + wm_SysKeyUp;
    procedure WMSysChar(var Msg: TMessage);
      virtual wm_First + wm_SysChar;
    procedure WMIMEReport(var Msg: TMessage);
      virtual wm_First + WM_IME_REPORT;
    procedure WMKeyDown(var Msg: TMessage);
      virtual wm_First + wm_KeyDown;
    procedure WMKeyUp(var Msg: TMessage);
      virtual wm_First + wm_KeyUp;
    procedure WMChar(var Msg: TMessage);
      virtual wm_First + wm_Char;
    procedure InitMenu(var Menu: HMenu);
    procedure InitMenuPopup(SubMenu: HMenu);
    procedure WMInitMenuPopup(var Msg: TMessage);
      virtual wm_First + wm_InitMenuPopup;
    procedure WMActivate(var Msg: TMessage);
      virtual wm_First + wm_Activate;
    procedure Paint(PaintDC: HDC; var PaintInfo:TPaintStruct); virtual;
    procedure ButtonDown(Msg: TMessage; LMR: word);
    procedure WMLButtonDown(Var Msg: TMessage);
      virtual wm_First + wm_LButtonDown;
    procedure ButtonUp(Paste: boolean);
    procedure WMLButtonUp(Var Msg: TMessage);
      virtual wm_First + wm_LButtonUp;
    procedure ButtonDblClk(Msg: TMessage);
    procedure WMLButtonDblClk(Var Msg: TMessage);
      virtual wm_First + wm_LButtonDblClk;
    procedure WMMButtonDown(Var Msg: TMessage);
      virtual wm_First + wm_MbuttonDown;
    procedure WMMButtonUp(Var Msg: TMessage);
      virtual wm_First + wm_MbuttonUp;
    procedure WMNCRButtonDown(Var Msg: TMessage);
      virtual wm_First + wm_NCRButtonDown;
    procedure WMRButtonDown(Var Msg: TMessage);
      virtual wm_First + wm_RButtonDown;
    procedure WMRButtonUp(Var Msg: TMessage);
      virtual wm_First + wm_RButtonUp;
    procedure WMMouseMove(Var Msg: TMessage);
      virtual wm_First + wm_MouseMove;
    procedure ChangeSelectRegion;

    procedure WMSize(Var Msg: TMessage);
      virtual wm_First + wm_Size;
    procedure WMNCLButtonDblClk(var Msg: TMessage);
      virtual wm_First + wm_NCLButtonDblClk;
    procedure WMMove(var Msg: TMessage);
      virtual wm_First + wm_Move;
    procedure WMTimer(var Msg: TMessage);
      virtual wm_First + wm_Timer;

    procedure WMAccelCommand(var Msg: TMessage);
      virtual wm_User + MsgAccelCommand;
    procedure WMChangeMenu(var Msg: TMessage);
      virtual wm_User + MsgChangeMenu;
    procedure WMCloseIME(var Msg: TMessage);
      virtual wm_User + MsgCloseIME;
    procedure WMComNotify(var Msg: TMessage);
      virtual wm_User + MsgComNotify;
    procedure WMComOpen(var Msg: TMessage);
      virtual wm_User + MsgComOpen;
    procedure WMComStart(var Msg: TMessage);
      virtual wm_User + MsgComStart;
    procedure WMDlgHelp(var Msg: TMessage);
      virtual wm_User + MsgDlgHelp2;

    procedure WMCommNotify(var Msg: TMessage);
      virtual wm_First + wm_CommNotify;
    procedure WMSysColorChange(var Msg: TMessage);
      virtual wm_First + wm_SysColorChange;
    procedure WMSysCommand(var Msg: TMessage);
      virtual wm_First + wm_SysCommand;

    procedure WMDdeEnd(var Msg: TMessage);
      virtual wm_User + MsgDdeEnd;

    procedure CMFileNewConnection(var Msg: TMessage);
      virtual cm_First + cm_FileNewConnection;
    procedure CMFileLog(var Msg: TMessage);
      virtual cm_First + cm_FileLog;
    procedure CMFileSend(var Msg: TMessage);
      virtual cm_First + cm_FileSend;
    procedure CMFileKermitRcv(var Msg: TMessage);
      virtual cm_First + cm_FileKermitRcv;
    procedure CMFileKermitGet(var Msg: TMessage);
      virtual cm_First + cm_FileKermitGet;
    procedure CMFileKermitSend(var Msg: TMessage);
      virtual cm_First + cm_FileKermitSend;
    procedure CMFileKermitFinish(var Msg: TMessage);
      virtual cm_First + cm_FileKermitFinish;
    procedure CMFileXRcv(var Msg: TMessage);
      virtual cm_First + cm_FileXRcv;
    procedure CMFileXSend(var Msg: TMessage);
      virtual cm_First + cm_FileXSend;
    procedure CMFileZRcv(var Msg: TMessage);
      virtual cm_First + cm_FileZRcv;
    procedure CMFileZSend(var Msg: TMessage);
      virtual cm_First + cm_FileZSend;
    procedure CMFileBPRcv(var Msg: TMessage);
      virtual cm_First + cm_FileBPRcv;
    procedure CMFileBPSend(var Msg: TMessage);
      virtual cm_First + cm_FileBPSend;
    procedure CMFileQVRcv(var Msg: TMessage);
      virtual cm_First + cm_FileQVRcv;
    procedure CMFileQVSend(var Msg: TMessage);
      virtual cm_First + cm_FileQVSend;

    procedure CMFileChangeDir(var Msg: TMessage);
      virtual cm_First + cm_FileChangeDir;
    procedure Print(MC: bool);
    procedure CMFilePrint(var Msg: TMessage);
      virtual cm_First + cm_FilePrint;
    procedure CMFileDisconnect(var Msg: TMessage);
      virtual cm_First + cm_FileDisconnect;

    procedure CMEditCopy(var Msg: TMessage);
      virtual cm_First + cm_EditCopy;
    procedure CMEditPaste(var Msg: TMessage);
      virtual cm_First + cm_EditPaste;
    procedure CMEditPasteCR(var Msg: TMessage);
      virtual cm_First + cm_EditPasteCR;
    procedure CMEditClearScreen(var Msg: TMessage);
      virtual cm_First + cm_EditClearScreen;
    procedure CMEditClearBuffer(var Msg: TMessage);
      virtual cm_First + cm_EditClearBuffer;

    procedure CMSetupTerminal(var Msg: TMessage);
      virtual cm_First + cm_SetupTerminal;
    procedure InitTerm;
    procedure CMSetupWindow(var Msg: TMessage);
      virtual cm_First + cm_SetupWindow;
    procedure InitWin;
    procedure CMSetupFont(var Msg: TMessage);
      virtual cm_First + cm_SetupFont;
    procedure CMSetupKeyboard(var Msg: TMessage);
      virtual cm_First + cm_SetupKeyboard;
    procedure CMSetupSerialPort(var Msg: TMessage);
      virtual cm_First + cm_SetupSerialPort;
    procedure CMSetupTCPIP(var Msg: TMessage);
      virtual cm_First + cm_SetupTCPIP;
    procedure CMSetupGeneral(var Msg: TMessage);
      virtual cm_First + cm_SetupGeneral;
    procedure CMSetupSave(var Msg: TMessage);
      virtual cm_First + cm_SetupSave;
    procedure ResetSetup;
    procedure CMSetupRestore(var Msg: TMessage);
      virtual cm_First + cm_SetupRestore;

    procedure CMControlResetTerminal(var Msg: TMessage);
      virtual cm_First + cm_ControlResetTerminal;
    procedure CMControlAreYouThere(var Msg: TMessage);
      virtual cm_First + cm_ControlAreYouThere;
    procedure CMControlSendBreak(var Msg: TMessage);
      virtual cm_First + cm_ControlSendBreak;
    procedure CMControlResetPort(var Msg: TMessage);
      virtual cm_First + cm_ControlResetPort;
    procedure CMControlOpenTEK(var Msg: TMessage);
      virtual cm_First + cm_ControlOpenTEK;
    procedure CMControlCloseTEK(var Msg: TMessage);
      virtual cm_First + cm_ControlCloseTEK;
    procedure CMControlMacro(var Msg: TMessage);
      virtual cm_First + cm_ControlMacro;

    procedure CMHelpIndex(var Msg: TMessage);
      virtual cm_First + cm_HelpIndex;
    procedure CMHelpUsing(var Msg: TMessage);
      virtual cm_First + cm_HelpUsing;
    procedure CMAbout(var Msg: TMessage);
      virtual cm_First + cm_About;

  end;

var
  VTPtr: PVTWindow;

implementation

constructor VTScroller.Init(TheWindow: PVTWindow; TheXUnit, TheYUnit: Integer;
                            TheXRange, TheYRange: Longint);
begin
  Window := TheWindow;
  Parent := TheWindow;
  TScroller.Init(TheWindow,TheXUnit,TheYUnit,TheXRange,TheYRange);
end;

procedure VTScroller.HScroll(ScrollRequest: Word; ThumbPos: Integer);
begin
  SetPageSize; {for safety}
  if XRange > 0 then
  begin
    TScroller.HScroll(ScrollRequest, ThumbPos);
    Parent^.WinOrgX := XPos;
    Parent^.NewOrgX := XPos;
    Parent^.CaretOn;
  end;
end;

procedure VTScroller.VScroll(ScrollRequest: Word; ThumbPos: Integer);
begin
  SetPageSize; {for safety}
  if YRange > 0 then
  begin
    TScroller.VScroll(ScrollRequest, ThumbPos);
    Parent^.WinOrgY := YPos - PageStart;
    Parent^.NewOrgY := YPos - PageStart;
    Parent^.CaretOn;
  end;
end;

constructor VTWindow.Init(Aparent: PWindowsObject);
var
 i: integer;
 Ny: word;
 Temp: array[0..256] of char;
 KeyCnfFN: array[0..fsPathName] of char;
begin
  MsgDlgHelp := RegisterWindowMessage(HELPMSGSTRING);

  StartTeraTerm(@ts);

  {Get command line}
  i := PByte(Ptr(GetCurrentPDB,$80))^;
  Move(PChar(Ptr(GetCurrentPDB,$81))[0],Temp[0],i);
  Temp[i] := #0;

  if LoadTTSET then
    ParseParam(Temp, @ts, TopicName);
  FreeTTSET;

  InitKeyboard;
  SetKeyMap;

  if ts.Dlg3d>0 then
    ts.Dlg3d := word(LoadCtl3d);

  TWindow.Init(Aparent,ts.Title);

  if ts.HideTitle>0 then
    Attr.Style := ws_Visible or ws_popup or ws_thickframe or
                  ws_border or ws_VScroll or ws_HScroll
  else
    Attr.Style := ws_Visible or ws_VScroll or ws_HScroll or
                  ws_Caption or ws_SysMenu or ws_MinimizeBox or
                  ws_Border or ws_ThickFrame;

  Attr.X := ts.VTPos.X;
  Attr.Y := ts.VTPos.Y;

  if (ts.HideTitle=0) and (ts.PopupMenu=0) then
    InitMenu(Attr.Menu)
  else
    Attr.Menu := 0;

  {setup terminal}
  NumOfColumns := ts.TerminalWidth;
  NumOfLines := ts.TerminalHeight;

  {setup window}
  if (ts.EnableScrollBuff>0) then
  begin
    if (ts.ScrollBuffMax < NumOfLines) then
      ts.ScrollBuffMax := NumOfLines;
    Ny := ts.ScrollBuffMax;
  end
  else
    Ny := NumOfLines;

  if not ChangeBuffer(NumOfColumns,Ny) then
    PostQuitMessage(0);

  if ts.EnableScrollBuff>0 then
    ts.ScrollBuffMax := NumOfLinesInBuff;

  {Initialize scroller}
  Scroller := New(PVTScroller, Init(@Self, 1, 1, 0, 1));
  Scroller^.XPage := NumOfColumns;
  Scroller^.XRange := 0;
  Scroller^.YPage := NumOfLines;
  Scroller^.YRange := 0;
  Scroller^.TrackMode := true;
  Scroller^.AutoOrg := false;

  AdjustSize := true;

  {Emulation mode}
  ActiveWin := IdVT;
  DebugFlag := false;

  {Keyboard status}
  PreviousKey := 0;

  {window status}
  Active := false;
  Minimized := false;
  DblClk := false;
  DC := 0;

  FirstPaint := true;
end;

procedure VTWindow.SetupWindow;
var
  TmpDC: HDC;
  i: integer;
begin
  HMainWin := HWindow;

  if not ts.Ok and LoadTTDLG then
  begin
    if SetupGeneral(HWindow,@ts.PortType,@ts.Language) then
    begin
      FreeTTDLG;
      if LoadTTSET then
      begin
        WriteIniFile(ts.SetupFName,@ts);
        FreeTTSET;
        ts.Ok := true;
        ChangeDefaultSet(@ts);
      end;
    end
    else
      FreeTTDLG;
    if not ts.Ok then PostQuitMessage(0);
  end;

  {Reset Terminal}
  ResetTerminal;

  TmpDC := GetDC(HWindow);

  ANSIColor[IdBack ]   := RGB(  0,  0,  0);
  ANSIColor[IdRed  ]   := RGB(255,  0,  0);
  ANSIColor[IdGreen]   := RGB(  0,255,  0);
  ANSIColor[IdYellow]  := RGB(255,255,  0);
  ANSIColor[IdBlue]    := RGB(  0,  0,255);
  ANSIColor[IdMagenta] := RGB(255,  0,255);
  ANSIColor[IdCyan]    := RGB(  0,255,255);
  ANSIColor[IdFore ]   := RGB(255,255,255);

  for i := IdBack to IdFore do
    ANSIColor[i] := GetNearestColor(TmpDC, ANSIColor[i]);

  {background paintbrush}
  BackGround := CreateSolidBrush(ts.VTColor[1]);
  {CRT width & height}
  CRTWidth := GetDeviceCaps(TmpDC,HORZRES);
  CRTHeight := GetDeviceCaps(TmpDC,VERTRES);
  ReleaseDC(HWindow, TmpDC);

  if (ts.VTPos.X > CRTWidth) or (ts.VTPos.Y > CRTHeight) then
  begin
    ts.VTPos.X := CW_UseDefault;
    ts.VTPos.Y := CW_UseDefault;
    Attr.X := ts.VTPos.X;
    Attr.Y := ts.VTPos.Y;
  end;

  if (ts.TEKPos.X > CRTWidth) or (ts.TEKPos.Y > CRTHeight) then
  begin
    ts.TEKPos.X := CW_UseDefault;
    ts.TEKPos.Y := CW_UseDefault;
  end;

  if ts.PopupMenu>0 then
    PostMessage(HWindow,wm_user+MsgChangeMenu,0,0);

  ChangeFont;

  ResetIME;

  TWindow.SetupWindow;

  ChangeWindowSize(NumOfColumns,NumOfLines);

  WinWidthOld := NumOfColumns;
  WinHeightOld := NumOfLines;

  ChangeTitle;

  {Enable drag-drop}
  DragAcceptFiles(HWindow,true);
end;

function VTWindow.GetClassName: PChar;
begin
  GetClassName := 'VTWin';
end;

procedure VTWindow.DefWndProc(var Msg: TMessage);
begin
  if Msg.Message=MsgDlgHelp then
    WMDlgHelp(Msg)
  else if (ts.HideTitle>0) and
          (Msg.Message=wm_first+wm_nchittest) then
  begin
    TWindow.DefWndProc(Msg);
    if (Msg.Result=HTCLIENT) and AltKey then
      Msg.Result := HTCAPTION;
  end
  else
    TWindow.DefWndProc(Msg);
end;

procedure VTWindow.GetWindowClass(var AWndClass: TWndClass);
begin
  TWindow.GetWindowClass(AWndClass);
  AWndClass.hbrBackground := 0;
  AWndClass.HIcon := LoadIcon(HInstance, PChar(IdVTIcon));
  AWndClass.HCursor := LoadCursor(0, idc_IBeam);
  AWndClass.style := AWndClass.style or cs_dblclks; {to receive wm_lbuttondblclk}
end;

procedure VTWindow.ResetTerminal; {reset variables but don't update screen}
var
  i: integer;
begin
  {Cursor}
  NewLine(PageStart);
  CursorX := 0;
  CursorY := 0;
  WinOrgX := 0;
  WinOrgY := 0;
  NewOrgX := 0;
  NewOrgY := 0;
  SavedCursor.X := 0;
  SavedCursor.Y := 0;

  {Top/bottom mergin}
  CursorTop := 0;
  CursorBottom := NumOfLines-1;

  {Initialize text selection region}
  SelectStart.X := 0;
  SelectStart.Y := 0;
  SelectEnd := SelectStart;
  SelectEndOld := SelectStart;
  Select := false;
  LButton := false;
  MButton := false;
  RButton := false;
  {Tab stops}
  NTabstops := (NumOfColumns-1) div 8;
  for i := 1 to NTabstops do
    TabStops[i-1] := i*8;

  {Attribute}
  CharAttr := AttrDefault;
  CharAttr2 := AttrDefault2;
  Special := false;

  {Various modes}
  AppliKeyMode := false;
  InsertMode := false;
  AppliCursorMode := false;
  RelativeOrgMode := false;
  AutoWrapMode := true;
  Wrap := false;
  ReverseColor := false;
  LFMode := ts.CRSend = IdCRLF;
  AutoRepeatMode := true;

  {Character sets}
  ResetCharSet;

  {ESC flag for device control sequence}
  ESCFlag := false;
  {for TEK sequence}
  JustAfterESC := false;

  {Parse mode}
  ParseMode := ModeFirst;

  StrChangeCount := 0;
  DCUseCount := 0;
  ScrollCount := 0;
  ScrollerChange := false;

  {Clear printer mode}
  PrinterMode := false;
  {IME}
  OpenConv := false;
end;

procedure VTWindow.ResetCharSet;
begin
  If ts.Language=IdJapanese then
  begin
    Gn[0] := IdASCII;
    Gn[1] := IdKatakana;
    Gn[2] := IdKatakana;
    Gn[3] := IdKanji;
    Glr[0] := 0;
    Glr[1] := 3;
  end
  else begin
    Gn[0] := IdASCII;
    Gn[1] := IdSpecial;
    Gn[2] := IdASCII;
    Gn[3] := IdSpecial;
    Glr[0] := 0;
    Glr[1] := 0;
    cv.SendCode := IdASCII;
    cv.SendKanjiFlag := false;
    cv.EchoCode := IdASCII;
    cv.EchoKanjiFlag := false;
  end;
  {Kanji flag}
  KanjiIn := false;
  EUCKanaIn := false;
  EUCsupIn := false;
  SSflag := false;

  cv.Language := ts.Language;
  cv.CRSend := ts.CRSend;
  cv.KanjiCodeEcho := ts.KanjiCode;
  cv.JIS7KatakanaEcho := ts.JIS7Katakana;
  cv.KanjiCodeSend := ts.KanjiCodeSend;
  cv.JIS7KatakanaSend := ts.JIS7KatakanaSend;
  cv.KanjiIn := ts.KanjiIN;
  cv.KanjiOut := ts.KanjiOut;
end;

procedure VTWindow.ResetIME;
begin
  {reset language for communication}
  cv.Language := ts.Language;

  {reset IME}
  if ts.Language=IdJapanese then
  begin
    if ts.UseIME=0 then
      FreeWINNLS
    else if not LoadWINNLS then
      ts.UseIME := 0;

    if (ts.UseIME=1) and
       (ts.IMEInline > 0) then
      CaretOn;
  end
  else
    FreeWINNLS;
end;

procedure VTWindow.ClearBuffer;
begin 
  {Reset buffer}
  ResetBuff;

  {Home position}
  CursorX := 0;
  CursorY := 0;
  WinOrgX := 0;
  WinOrgY := 0;
  NewOrgX := 0;
  NewOrgY := 0;

  {Top/bottom mergin}
  CursorTop := 0;
  CursorBottom := NumOfLines-1;

  InvalidateRect(HWindow,nil,false);

  ScrollCount := 0;
  StrChangeCount := 0;
  if WinHeight > NumOfLines then
    ChangeWindowSize(NumOfColumns,NumOfLines)
  else begin
    if (NumOfLines=WinHeight) and (ts.EnableScrollBuff>0)then
    begin
      Scroller^.SetRange(Scroller^.XRange,1);
      Scroller^.YRange := 0;
    end
    else
      Scroller^.SetRange(Scroller^.XRange,NumOfLines-WinHeight);

    Scroller^.ScrollTo(0,0);
  end;

  LockBuffer; {for safety}
end;

procedure VTWindow.ChangeCaret;
begin
  DestroyCaret;
  case ts.CursorShape of
    IdVCur: CreateCaret(HWindow, 0, CurWidth, FontHeight);
    IdHCur: CreateCaret(HWindow, 0, FontWidth, CurWidth);
  end;
  CaretStatus := 1;
  CaretOn;
end;

procedure VTWindow.CaretOn;
var
  CaretX, CaretY, H: integer;
begin
  if not Active then exit;

  CaretX := (CursorX-WinOrgX)*FontWidth;
  CaretY := (CursorY-WinOrgY)*FontHeight;

  if (ts.Language=IdJapanese) and
     (HWINNLS>=32) and (ts.IMEInline>0) then
    {set IME conversion window pos. & font}
    SetConversionWindow(HWindow,CaretX,CaretY,VTFont[0]);

  if ts.CursorShape<>IdVCur then
  begin
    if ts.CursorShape=IdHCur then
    begin
     CaretY := CaretY+FontHeight-CurWidth;
     H := CurWidth;
    end
    else H := FontHeight;

    DestroyCaret;
    LockBuffer;
    NewLine(PageStart+CursorY);
    if byte(AttrLine[CursorX]) and AttrKanji <> 0 then {DBCS first?}
      CreateCaret(HWindow, 0, FontWidth*2, H) {double width caret}
    else
      CreateCaret(HWindow, 0, FontWidth, H); {single width caret}
    UnlockBuffer;
    CaretStatus := 1;
  end;

  SetCaretPos(CaretX,CaretY);

  while CaretStatus > 0 do
  begin
    ShowCaret(HWindow);
    Dec(CaretStatus);
  end;

end;

procedure VTWindow.CaretOff;
begin
  if not Active then exit;
  if CaretStatus = 0 then
  begin
    HideCaret(HWindow);
    inc(CaretStatus);
  end;
end;

procedure VTWindow.Scroll1Line;
var
  i: integer;
  SrcPtr, DestPtr: longint;
begin
  UpdateStr;

  if (CursorTop = 0) and (CursorBottom = NumOfLines-1) then
  begin
    Dec(WinOrgY);
    if BufferFull then Dec(Scroller^.YPos);
    BuffScroll(1);
    CountScroll;
    ScrollerChange := true;
  end
  else
    if (CursorTop<=CursorY) and (CursorY<=CursorBottom) then
    begin
      DestPtr := GetLinePtr(PageStart+CursorTop);
      for i:= CursorTop+1 to CursorBottom do
      begin
        SrcPtr := NextLinePtr(DestPtr);
        Move(CodeBuff[SrcPtr],CodeBuff[DestPtr],NumOfColumns);
        Move(AttrBuff[SrcPtr],AttrBuff[DestPtr],NumOfColumns);
        Move(AttrBuff2[SrcPtr],AttrBuff2[DestPtr],NumOfColumns);
        DestPtr := SrcPtr;
      end;
      FillChar(CodeBuff[SrcPtr],NumOfColumns,$20);
      FillChar(AttrBuff[SrcPtr],NumOfColumns,AttrDefault);
      FillChar(AttrBuff2[SrcPtr],NumOfColumns,AttrDefault2);

      if Active and CompletelyVisible and (CursorBottom+1-WinOrgY <= WinHeight) then
      begin
        R.left := 0;
        R.right := ScreenWidth;
        R.top := (CursorTop-WinOrgY)*FontHeight;
        R.bottom := (CursorBottom+1-WinOrgY)*FontHeight;
        ScrollWindowEx(HWindow,0,-FontHeight,@R,@R,0,nil,0);
        R.top := R.bottom - FontHeight;
        FillRect(DC,R,BackGround);
      end
      else
        UpdateRect(DC,WinOrgX,CursorTop,WinOrgX+WinWidth-1,CursorBottom);

    end;
end;

procedure VTWindow.SetLogFont;
begin
  FillChar(VTLogFont, SizeOf(TLogFont), #0);
  with VTLogFont do
  begin
    lfWeight := FW_Normal;
    lfItalic := 0;
    lfUnderLine := 0;
    lfStrikeOut := 0;
    lfWidth := ts.VTFontSize.X;
    lfHeight := ts.VTFontSize.Y;
    lfCharset := ts.VTFontCharSet;

    lfOutPrecision  := Out_Character_Precis;
    lfClipPrecision := Clip_Character_Precis;
    lfQuality       := Default_Quality;
    lfPitchAndFamily:= Fixed_Pitch or FF_Modern;
    StrCopy(lfFacename,ts.VTFont);
  end;
end;

procedure VTWindow.ChangeFont;
var
  i, j: integer;
  Metrics: TTextMetric;
  TmpDC: HDC;
begin
  {Delete Old Fonts}
  for i := 0 to AttrFontMask do
  begin
    for j := i+1 to AttrFontMask do
      if VTFont[j]=VTFont[i] then
        VTFont[j] := 0;
    if VTFont[i]<>0 then DeleteObject(VTFont[i]);
  end;

  {Normal Font}
  SetLogFont;
  VTFont[0] := CreateFontIndirect(VTLogFont);

  TmpDC := GetDC(HWindow);

  SelectObject(TmpDC, VTFont[0]);
  GetTextMetrics(TmpDC, Metrics); 
  FontWidth := Metrics.tmAveCharWidth + ts.FontDW;
  FontHeight := Metrics.tmHeight + ts.FontDH;

  ReleaseDC(HWindow,TmpDC);

  {Underline}
  VTLogFont.lfUnderLine := 1;
  VTFont[AttrUnder] := CreateFontIndirect(VTLogFont);

  if ts.EnableBold>0 then
  begin
    {Bold}
    VTLogFont.lfUnderLine := 0;
    VTLogFont.lfWeight := FW_Bold;
    VTFont[AttrBold] := CreateFontIndirect(VTLogFont);
    {Bold + Underline}
    VTLogFont.lfUnderLine := 1;
    VTFont[AttrBold or AttrUnder] := CreateFontIndirect(VTLogFont);
  end
  else begin
    VTFont[AttrBold] := VTFont[AttrDefault];
    VTFont[AttrBold or AttrUnder] := VTFont[AttrUnder];
  end;
  {Special font}                 
  VTLogFont.lfWeight := FW_Normal;
  VTLogFont.lfUnderLine := 0;
  VTLogFont.lfWidth := FontWidth + 1; {adjust width}
  VTLogFont.lfHeight := FontHeight;
  VTLogFont.lfCharset := 1;
  VTLogFont.lfPitchAndFamily:= DEFAULT_PITCH or FF_DONTCARE;

  StrCopy(VTLogFont.lfFacename,'Tera Special');
  VTFont[AttrSpecial] := CreateFontIndirect(VTLogFont);
  VTFont[AttrSpecial or AttrBold] := VTFont[AttrSpecial];
  VTFont[AttrSpecial or AttrUnder] := VTFont[AttrSpecial];
  VTFont[AttrSpecial or AttrBold or AttrUnder] := VTFont[AttrSpecial];

  SetLogFont;

  for i := 0 to 255 do
    Dx[i] := FontWidth;

  Scroller^.SetUnits(FontWidth, FontHeight);

end;

procedure VTWindow.DispChar(b: byte);
var
  SpecialNew: boolean;
  XStart: integer;
  CharAttrTmp: byte;
begin
  if Wrap then
  begin
    CarriageReturn;
    LineFeed(LF);
  end;
  if cv.HLogBuf<>0 then Log1Byte(b);
  Wrap := false;

  if ts.Language=IdJapanese then
  begin
    EraseKanji(1); {if cursor is on right half of a kanji, erase the kanji}
    if not InsertMode then EraseKanji(0); {if cursor on left half...}
  end;

  SpecialNew := false;
  if (b>$5F) and (b<$80) then
  begin
    if SSFlag then
      SpecialNew := Gn[Gltmp]=IdSpecial
    else
      SpecialNew :=  Gn[Glr[0]]=IdSpecial;
  end
  else if b>$DF then
  begin
    if SSFlag then
      SpecialNew := Gn[Gltmp]=IdSpecial
    else
      SpecialNew :=  Gn[Glr[1]]=IdSpecial;
  end;

  if SpecialNew xor Special then
  begin
    UpdateStr;
    Special := SpecialNew;
    SetupDC;
  end;

  if Special then
  begin
    b := b and $7F;
    CharAttrTmp := CharAttr or AttrSpecial;
  end
  else
    CharAttrTmp := CharAttr;

  if InsertMode
  then begin
    Move(CodeLine[CursorX],CodeLine[CursorX+1],NumOfColumns-1-CursorX);
    Move(AttrLine[CursorX],AttrLine[CursorX+1],NumOfColumns-1-CursorX);
    Move(AttrLine2[CursorX],AttrLine2[CursorX+1],NumOfColumns-1-CursorX);
    CodeLine[CursorX] := char(b);
    AttrLine[CursorX] := char(CharAttrTmp);
    AttrLine2[CursorX] := char(CharAttr2);
    {last char in current line is kanji first?}
    if (byte(AttrLine[NumOfColumns-1]) and AttrKanji) <> 0 then
    begin
      {then delete it}
      CodeLine[NumOfColumns-1] := #$20;
      AttrLine[NumOfColumns-1] := char(AttrDefault);
      AttrLine2[NumOfColumns-1] := char(AttrDefault2);
    end;

    if StrChangeCount=0 then Xstart := CursorX
                        else Xstart := StrChangeStart;
    StrChangeCount := 0;
    UpdateRect(DC,XStart,CursorY,NumOfColumns-1,CursorY);
  end
  else begin
    CodeLine[CursorX] := char(b);
    AttrLine[CursorX] := char(CharAttrTmp);
    AttrLine2[CursorX] := char(CharAttr2);

    if StrChangeCount=0 then StrChangeStart := CursorX;
    inc(StrChangeCount);
  end;

  if CursorX < NumOfColumns-1
    then MoveRight
    else begin
      UpdateStr;
      Wrap := AutoWrapMode;
    end;

end;

procedure VTWindow.DispKanji(b: byte);
var
  XStart: integer;
begin

  Kanji := Kanji + b;

  if ConvJIS then
    Kanji := JIS2SJIS(Kanji and $7f7f);

  if Wrap then
  begin
    CarriageReturn;
    LineFeed(LF);
  end
  else if (CursorX > NumOfColumns-2) then
    if AutowrapMode then
    begin
      CarriageReturn;
      LineFeed(LF);
    end
    else exit;

  Wrap := false;

  if cv.HLogBuf<>0 then
  begin
    Log1Byte(Hi(Kanji));
    Log1Byte(Lo(Kanji));
  end;

  EraseKanji(1); {if cursor is on right half of a kanji, erase the kanji}

  if Special then
  begin
    UpdateStr;
    Special := false;
    SetupDC;
  end;

  if InsertMode
  then begin
    Move(CodeLine[CursorX],CodeLine[CursorX+2],NumOfColumns-2-CursorX);
    Move(AttrLine[CursorX],AttrLine[CursorX+2],NumOfColumns-2-CursorX);
    Move(AttrLine2[CursorX],AttrLine2[CursorX+2],NumOfColumns-2-CursorX);

    CodeLine[CursorX] := char(Hi(Kanji));
    AttrLine[CursorX] := char(CharAttr or AttrKanji); {DBCS first byte}
    AttrLine2[CursorX] := char(CharAttr2);
    if CursorX < NumOfColumns-1 then
    begin
      CodeLine[CursorX+1] := char(Lo(Kanji));
      AttrLine[CursorX+1] := char(CharAttr);
      AttrLine2[CursorX+1] := char(CharAttr2);
    end;

    {last char in current line is kanji first?}
    if (byte(AttrLine[NumOfColumns-1]) and AttrKanji) <> 0 then
    begin
      {then delete it}
      CodeLine[NumOfColumns-1] := #$20;
      AttrLine[NumOfColumns-1] := char(AttrDefault);
      AttrLine2[NumOfColumns-1] := char(AttrDefault2);
    end;

    if StrChangeCount=0 then Xstart := CursorX
                        else Xstart := StrChangeStart;
    StrChangeCount := 0;
    UpdateRect(DC,XStart,CursorY,NumOfColumns-1,CursorY);

  end
  else begin
    CodeLine[CursorX] := char(Hi(Kanji));
    AttrLine[CursorX] := char(CharAttr or AttrKanji); {DBCS first byte}
    AttrLine2[CursorX] := char(CharAttr2);
    if CursorX < NumOfColumns-1 then
    begin
      CodeLine[CursorX+1] := char(Lo(Kanji));
      AttrLine[CursorX+1] := char(CharAttr);
      AttrLine2[CursorX+1] := char(CharAttr2);
    end;

    if StrChangeCount=0 then StrChangeStart := CursorX;
    StrChangeCount := StrChangeCount + 2;
  end;

  if CursorX < NumOfColumns-2
  then begin
    MoveRight;
    MoveRight;
  end
  else begin
    UpdateStr;
    Wrap := AutoWrapMode;
  end;

end;

procedure VTWindow.MoveRight;
{ move cursor right, but dont update screen.
  this procedure must be called from DispChar&DispKanji only}
begin
  inc(CursorX);

  ScrollToCursor;
end;

procedure VTWindow.EraseKanji(LR: integer);
begin
  if (CursorX-LR>=0) and
     ((byte(AttrLine[CursorX-LR]) and AttrKanji) <> 0) then
  begin
    CodeLine[CursorX-LR] := #$20;
    AttrLine[CursorX-LR] := char(AttrDefault);
    AttrLine2[CursorX-LR] := char(AttrDefault2);
    if CursorX-LR+1 < NumOfColumns then
    begin
      CodeLine[CursorX-LR+1] := #$20;
      AttrLine[CursorX-LR+1] := char(AttrDefault);
      AttrLine2[CursorX-LR+1] := char(AttrDefault2);
    end;
  end
end;

procedure VTWindow.BackSpace;
begin
  if CursorX > 0 then
  begin
    MoveCursor(CursorX-1,CursorY);
    if cv.HLogBuf<>0 then Log1Byte(BS);
  end;
end;

procedure VTWindow.CarriageReturn;
begin
 if (cv.HLogBuf<>0) and (CursorX>0) then Log1Byte(CR);

 MoveCursor(0,CursorY);
end;

procedure VTWindow.PrintCurrentLine(b: byte);
var
  i: integer;
begin
  if HPrnFile<=0 then exit;
  i := NumOfColumns;
  while (i>0) and (CodeLine[i-1]=#$20) do
    dec(i);
  _lwrite(HPrnFile,CodeLine,i);

  if (b>=LF) and (b<=FF) then
  begin
    _lwrite(HPrnFile,#$0D,1);
    _lwrite(HPrnFile,@b,1);
  end;
end;

procedure VTWindow.LineFeed(b: byte);
begin
 {for auto print mode}
 if AutoPrintMode and
    (b>=LF) and (b<=FF) then
   PrintCurrentLine(b);

 if cv.HLogBuf<>0 then Log1Byte(LF);

 if CursorY < CursorBottom then
   MoveCursor(CursorX,CursorY+1)
 else if CursorY = CursorBottom then Scroll1Line
 else if CursorY < NumOfLines-1 then
   MoveCursor(CursorX,CursorY+1);

 if LFMode then CarriageReturn;
end;

procedure VTWindow.Tab;
var
  i: integer;
begin
  if NTabStops>0
  then begin
    i := -1;
    repeat
      inc(i);
    until (TabStops[i]>CursorX) or (i>=NTabStops-1);
    if TabStops[i]>CursorX then MoveCursor(TabStops[i],CursorY)
                           else MoveCursor(NumOfColumns-1,CursorY);
  end
  else
    MoveCursor(NumOfColumns-1,CursorY);
  if cv.HLogBuf<>0 then Log1Byte(HT);
end;

procedure VTWindow.DispAllChar(b: integer);
begin
  InsertMode := false;
  AutoWrapMode := true;

  if (b and $80) = $80 then
  begin
    UpdateStr;
    CharAttr := AttrReverse;
    SetupDC;
    b := b and $7f;
  end;

  case b of
    NUL..US: begin
               DispChar(Ord('^'));
               DispChar(b+$40);
             end;
    DEL: begin
           DispChar(Ord('<'));
           DispChar(Ord('D'));
           DispChar(Ord('E'));
           DispChar(Ord('L'));
           DispChar(Ord('>'));
         end;
  else
    DispChar(b);
  end;

  if CharAttr <> AttrDefault then
  begin
    UpdateStr;
    CharAttr := AttrDefault;
    SetupDC;
  end;

end;

function VTWindow.Parse: integer;
var
  DummyMsg: TMsg;
  b: byte;
  c: integer;
begin
  c := CommRead1Byte(@cv,@b);

  if (c=0) or LButton or MButton or RButton then
  begin
    Parse := 0;
    exit
  end;

  CaretOff;

  ChangeEmu := 0;

  {Get Device Context}
  DC := GetDC(HWindow);
  while DC=0 do
  begin
    PeekMessage(DummyMsg, 0, 0, 0, pm_NoRemove);
    DC := GetDC(HWindow);
  end;
  SetupDC;
  DCUseCount := 0;

  LockBuffer;
  NewLine(PageStart+CursorY);

  while (c>0) and (ChangeEmu=0) do
  begin
    if DebugFlag then DispAllChar(b)
    else if PrinterMode then
      PrnParse(b)
    else begin

      if not BuffLock then {for safety}
      begin
        LockBuffer;
        NewLine(PageStart+CursorY);
      end;

      case ParseMode of
        ModeFirst: ParseFirst(b);
        ModeESC: EscapeSequence(b);
        ModeDCS: DeviceControl(b);
        ModeDCUserKey: DCUserKey(b);
        ModeSOS: IgnoreString(b);
        ModeCSI: ControlSequence(b);
        ModeXS:  XSequence(b);
        ModeDLE: DLESeen(b);
        ModeCAN: CANSeen(b);
      else
        begin
          ParseMode := ModeFirst;
          ParseFirst(b);
        end;
      end;

    end;

    if (ChangeEmu=0) then
      c := CommRead1Byte(@cv,@b);
  end;

  if StrChangeCount>0 then UpdateStr;

  UpdateScroll;

  UnlockBuffer;

  ReleaseDC(HWindow,DC);
  DC := 0;

  CaretOn;

  if ChangeEmu > 0 then ParseMode := ModeFirst;
  Parse := ChangeEmu;
end;

procedure VTWindow.ParseFirst(b: byte);
begin
  if (ts.Language=IdJapanese) and KanjiIn then
  begin
    if ((not ConvJIS) and ($3F<b) and (b<$FD) or
       ConvJIS and ( ($20<b) and (b<$7f) or ($a0<b) and (b<$ff) ))
    then begin
      DispKanji(b);
      KanjiIn := false;
      exit;
    end
    else
      if not ts.CtrlInKanji then
        KanjiIn := false;
  end;
        
  if SSFlag then
  begin
    if ts.Language=IdJapanese then
    begin
      if Gn[GLTmp] = IdKanji then
      begin
        Kanji := b shl 8;
        KanjiIn := true;
        SSFlag := false;
        exit;
      end
      else if Gn[GLtmp] = IdKatakana then b := b or $80;
    end;    
    DispChar(b);
    SSFlag := false;
    exit;
  end;

  if (not EUCsupIn) and (not EUCKanaIn) and
     (not KanjiIn) and CheckKanji(b) then
  begin
    Kanji := b shl 8;
    KanjiIn := true;
    exit;
  end;

  case b of
    NUL..US: ParseControl(b);
    $20: DispChar(b);
    $21..$7E: begin
        if ts.Language=IdJapanese then
        begin
          if EUCsupIn then
          begin
            dec(EUCcount);
            EUCsupIn := EUCcount=0;
            exit;
          end;

          if (Gn[Glr[0]] = IdKatakana) or EUCKanaIn then
          begin
            b := b or $80;
            EUCKanaIn := false;
          end;
        end;
        DispChar(b);
      end;
    $80..$8D: ParseControl(b);
    $8E: if (ts.Language=IdJapanese) and
            (ts.KanjiCode=IdEUC) then
           EUCKanaIn := true
         else
           ParseControl(b);
    $8F: if (ts.Language=IdJapanese) and
            (ts.KanjiCode=IdEUC) then
         begin
           EUCcount := 2;
           EUCsupIn := true;
         end else
           ParseControl(b);
    $90..$9F: ParseControl(b);
    $A0: DispChar($20);
    $A1..$FE: begin
      if ts.Language=IdJapanese then
      begin
        if EUCsupIn then
        begin
          dec(EUCcount);
          EUCsupIn := EUCcount=0;
          exit;
        end;
        EUCKanaIn := false;

        if (ts.KanjiCode=IdJIS) then
        begin
          if (ts.JIS7Katakana=1) and
             (Gn[Glr[1]] = IdASCII) then
            b := b and $7f;
        end;
      end;
      DispChar(b);
    end;
  end;

end;

procedure VTWindow.SetupDC;
var
  TextColor, BackColor: TColorRef;
  IFont: integer;
begin
  IFont := CharAttr and AttrFontMask;
  if Special then IFont := IFont or AttrSpecial;
  SelectObject(DC, VTFont[IFont]);

  if CharAttr and AttrBlink <> 0 then
    TextColor := ts.VTBlinkColor[0]
  else if CharAttr and AttrBold <> 0 then
    TextColor := ts.VTBoldColor[0]
  else if CharAttr2 and Attr2Fore <> 0 then
    TextColor := ANSIColor[CharAttr2 and Attr2ForeMask]
  else
    TextColor := ts.VTColor[0];

  if CharAttr and AttrBlink <> 0 then
    BackColor := ts.VTBlinkColor[1]
  else if CharAttr and AttrBold <> 0 then
    BackColor := ts.VTBoldColor[1]
  else if CharAttr2 and Attr2Back <> 0 then
    BackColor := ANSIColor[CharAttr2 and Attr2BackMask shr SftAttrBack]
  else
    BackColor := ts.VTColor[1];

  if CharAttr and AttrReverse <> 0 then
  begin
    SetTextColor(DC,BackColor);
    SetBkColor(DC,TextColor);
  end
  else begin
    SetTextColor(DC,TextColor);
    SetBkColor(  DC,BackColor);
  end;

  SetBkMode(DC,2); {OPAQUE}
end;

procedure VTWindow.CountDCUse;
var
  DummyMsg: TMsg;
begin
  inc(DCUseCount);
  if DCUseCount>=DCUseCountLimit then
  begin
    UnlockBuffer;
    DCUseCount := 0;
    ReleaseDC(HWindow,DC);
    DC := 0;
    while DC=0 do
    begin
      PeekMessage(DummyMsg, 0, 0, 0, pm_NoRemove);
      DC := GetDC(HWindow);
    end;
    SetupDC;

    LockBuffer;
    NewLine(PageStart+CursorY);
  end;
end;

procedure VTWindow.CountScroll;
begin
  inc(ScrollCount);
  if ScrollCount>=ts.ScrollThreshold then UpdateScroll;
end;

procedure VTWindow.UpdateScroll;
begin
  if (NewOrgX<>WinOrgX) or
     (NewOrgY<>WinOrgY) then
  begin
    if BuffEnd-WinHeight > 0 then
    begin
      Scroller^.YRange := BuffEnd-WinHeight;
      SetScrollRange(HWindow,SB_VERT,0,Scroller^.YRange,false);
    end;
    Scroller^.ScrollTo(NewOrgX,NewOrgY+PageStart);
  end
  else if ScrollerChange then
  begin
    Scroller^.YRange := BuffEnd-WinHeight;
    SetScrollRange(HWindow,SB_VERT,0,Scroller^.YRange,false);
    SetScrollPos(HWindow,SB_VERT,Scroller^.YPos,true);
  end;

  ScrollerChange := false;
  ScrollCount := 0;
end;

procedure VTWindow.MoveCursor(Xnew,Ynew: integer);
begin
  UpdateStr;

  if (CursorY<>Ynew) then NewLine(PageStart+Ynew);

  CursorX := Xnew;
  CursorY := Ynew;
  Wrap := false;

  ScrollToCursor;
end;

procedure VTWindow.UpdateStr;
begin
  if StrChangeCount=0 then exit;
  if (CursorY<WinOrgY) or
     (CursorY>=WinOrgY+WinHeight) then
  begin
    StrChangeCount := 0;
    exit;
  end;

  R.left := (StrChangeStart-WinOrgX)*FontWidth;
  R.right := R.left + StrChangeCount*FontWidth;
  R.top := (CursorY-WinOrgY)*FontHeight;
  R.bottom := R.top + FontHeight;

  ExtTextOut(DC,R.left+ts.FontDX,R.top+ts.FontDY,
             6,@R,@CodeLine[StrChangeStart],StrChangeCount,@Dx);
  StrChangeCount := 0;
  CountDCUse;
end;

procedure VTWindow.UpdateRect(UpdateDC: HDC; XStart,YStart,XEnd,YEnd: integer);
var
  i, j, count: integer;
  IStart, IEnd: integer;
  X, Y: integer;
  TmpPtr: Longint;
  CurAttr, TempAttr: byte;
  CurAttr2, TempAttr2: byte;
  CurSel, TempSel: boolean;
  RText: TRect;
  PrevFont: HFont;
  TextColor, BackColor: TColorRef;

  function CheckSelect(x,y: integer): boolean;
  var
    L, L1, L2: longint;
  begin
    L := makelong(x,y);
    L1 := makelong(SelectStart.X,SelectStart.Y);
    L2 := makelong(SelectEnd.X,SelectEnd.Y);

    CheckSelect := Select and
                   ((L1<=L) and (L<L2) or (L2<=L) and (L<L1))
  end;

begin

  if XStart >= WinOrgX+WinWidth then exit;
  if YStart >= WinOrgY+WinHeight then exit;
  if XEnd < WinOrgX then exit;
  if YEnd < WinOrgY then exit;

  if XStart < WinOrgX then XStart := WinOrgX;
  if YStart < WinOrgY then YStart := WinOrgY;
  if XEnd >= WinOrgX+WinWidth then XEnd := WinOrgX+WinWidth-1;
  if YEnd >= WinOrgY+WinHeight then YEnd := WinOrgY+WinHeight-1;

  PrevFont := SelectObject(UpdateDC, VTFont[0]);
  SetTextColor(UpdateDC, ts.VTColor[0]);
  SetBkColor(UpdateDC, ts.VTColor[1]);

  TempAttr := AttrDefault;
  TempAttr2 := AttrDefault2;
  TempSel := false;

  Y := (YStart-WinOrgY)*FontHeight;
  TmpPtr := GetLinePtr(PageStart+YStart);
  for j := YStart+PageStart to YEnd+PageStart do
  begin
    IStart := XStart;
    IEnd := XEnd;

    if (IStart>0) and
       (byte(AttrBuff[TmpPtr+IStart-1]) and AttrKanji <> 0) {DBCS first byte?}
    then Dec(IStart);

    X := (IStart-WinOrgX)*FontWidth;
    RText.top := Y;
    RText.bottom := Y + FontHeight;

    i := IStart;
    repeat
      CurAttr := byte(AttrBuff[TmpPtr+i]) and not AttrKanji;
      CurAttr2 := byte(AttrBuff2[TmpPtr+i]);
      CurSel := CheckSelect(i,j);
      count := 1;
      while (i+count <= IEnd) and
            (CurAttr=byte(AttrBuff[TmpPtr+i+count]) and not AttrKanji) and
            (CurAttr2=byte(AttrBuff2[TmpPtr+i+count])) and
            (CurSel=CheckSelect(i+count,j))   or
            (byte(AttrBuff[TmpPtr+i+count-1]) and AttrKanji <> 0) do
        Inc(count);
        
      if (CurAttr <> TempAttr) or
         (CurAttr2 <> TempAttr2) or
         (CurSel xor TempSel) then
      begin
        SelectObject(UpdateDC, VTFont[CurAttr and AttrFontMask]);

        if CurAttr and AttrBlink <> 0 then
          TextColor := ts.VTBlinkColor[0]
        else if CurAttr and AttrBold <> 0 then
          TextColor := ts.VTBoldColor[0]
        else if CurAttr2 and Attr2Fore <> 0 then
          TextColor := ANSIColor[CurAttr2 and Attr2ForeMask]
        else
          TextColor := ts.VTColor[0];

        if CurAttr and AttrBlink <> 0 then
          BackColor := ts.VTBlinkColor[1]
        else if CurAttr and AttrBold <> 0 then
          BackColor := ts.VTBoldColor[1]
        else if CurAttr2 and Attr2Back <> 0 then
          BackColor := ANSIColor[CurAttr2 and Attr2BackMask shr SftAttrBack]
        else
          BackColor := ts.VTColor[1];

        if CurSel xor (CurAttr and AttrReverse <> 0) then
        begin
          SetTextColor(UpdateDC,BackColor);
          SetBkColor(  UpdateDC,TextColor);
        end
        else begin
          SetTextColor(UpdateDC,TextColor);
          SetBkColor(  UpdateDC,BackColor);
        end;
        TempAttr := CurAttr;
        TempAttr2 := CurAttr2;
        TempSel := CurSel;
      end;
      RText.left := X;
      RText.right := X + count*FontWidth;
      ExtTextOut(UpdateDc,X+ts.FontDX,Y+ts.FontDY,
                 6,@RText,@CodeBuff[TmpPtr+i],count,@Dx);

      X := RText.right;
      i := i+count;
    until  (i>IEnd);
    Y := Y + FontHeight;
    TmpPtr := NextLinePtr(TmpPtr);
  end;
  SelectObject(UpdateDc, PrevFont);
  if UpdateDC=DC then SetupDC;
end;

procedure VTWindow.ScrollToCursor;
begin
  if CursorX < NewOrgX then
    NewOrgX := CursorX
  else if CursorX >= NewOrgX+WinWidth then
    NewOrgX := CursorX + 1 - WinWidth;

  if CursorY < NewOrgY then
    NewOrgY := CursorY
  else if CursorY >= NewOrgY+WinHeight then
    NewOrgY := CursorY + 1 - WinHeight;
end;

function VTWindow.CheckKanji(b: byte): boolean;
var
  Check: boolean;
begin
  CheckKanji := false;
  if ts.Language<>IdJapanese then exit;
  Check := false;
  case b of
    $21..$7E: Check := Gn[Glr[0]]=IdKanji;
    $A1..$FE: Check := (ts.KanjiCode = IdEUC) or
                       (ts.KanjiCode = IdJIS) and
                       (ts.JIS7Katakana = 1)  and
                       (Gn[Glr[1]]=IdKanji); 
  end;
  if Check then ConvJIS := true;

  if (ts.KanjiCode = IdSJIS) and
     (($80<b)and(b<$a0) or ($df<b)and(b<$fd)) then
  begin
    Check := true;
    ConvJIS := false;
  end;

  CheckKanji := Check;
end;

{$I vtesc.pas}
{$I vtwm.pas}
{$I vtcm.pas}

begin
  VTPtr := nil;
end.
